import logging

from hutch_python.cache import LoadCache
from hutch_python.load_conf import default_class_namespace
from hutch_python.utils import extract_objs

logger = logging.getLogger(__name__)


def test_load_cache_integration():
    logger.debug('test_load_cache_integration')
    cache = LoadCache('fake.db')
    cache(obj1=1, obj2=2, obj3=3)
    objs = extract_objs()
    assert objs['obj1'] == 1
    assert objs['obj2'] == 2
    assert objs['obj3'] == 3
    default_class_namespace(int, 'nums', cache)
    assert cache.objs.nums.obj1 == 1
    assert cache.objs.nums.obj2 == 2
    assert cache.objs.nums.obj3 == 3


def test_load_cache_importable():
    logger.debug('test_load_cache_importable')
    cache = LoadCache('fake2.db')
    cache(one=1)

    import hutch_python.db
    assert hutch_python.db.one == 1



"""
This module modifies an ``ipython`` shell to log inputs, outputs, and
tracebacks to a custom ``logger.input`` level. The ``INPUT`` level is lower
than the ``DEBUG`` level to avoid a terminal echo in debug mode.
"""
import functools
import logging
import sys
import textwrap
import threading
import traceback

from .constants import INPUT_LEVEL
from .log_setup import log_exception_to_central_server

logger = logging.getLogger(__name__)
logger.input = functools.partial(logger.log, INPUT_LEVEL)
logger.setLevel(INPUT_LEVEL)

_ip_logger = None


def _log_errors(func):
    """Decorator: wrap ``func`` to log exceptions."""

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            logger.input('Logging error', exc_info=True)
    return wrapped


_indented = functools.partial(textwrap.indent, prefix=' ' * 4)


class IPythonLogger:
    """
    Class that logs the most recent inputs, outputs, and exceptions at the
    custom ``INPUT`` level.

    Parameters
    ----------
    ipython : ``IPython.terminal.interactiveshell.TerminalInteractiveShell``
        The active ``ipython`` ``Shell``, perhaps the one returned by
        ``IPython.get_ipython()``.

    Attributes
    ----------
    ipython_in : list of str
        The IPython user input list.

    prev_err_value : Exception or None
        The last exception value that was logged.  Used for exception
        deduplication.

    line_in_progress : bool
        True if a line is currently in the process of being evaluated.
    """
    def __init__(self, ipython):
        self.ipython = ipython
        self.ipython_in = ipython.user_ns["In"]
        self.prev_err_value = None
        self.line_in_progress = False

        logging.addLevelName('INPUT', INPUT_LEVEL)

        ipython.events.register('pre_run_cell', self.log_user_input)
        ipython.events.register('post_run_cell', self.log_output)

        self._orig_sys_excepthook = sys.excepthook
        sys.excepthook = self._sys_exception_hook

        if hasattr(threading, "excepthook"):
            self._orig_thread_excepthook = threading.excepthook
            threading.excepthook = self._thread_excepthook

    def _should_log(self, exc_value):
        """Should the given exception be logged?"""
        return (
            isinstance(exc_value, Exception) and
            exc_value != self.prev_err_value
        )

    @_log_errors
    def _sys_exception_hook(self, exc_type, exc_value, exc_tb):
        """Unhandled exception hook - outside of input-related ones."""
        try:
            if self._should_log(exc_value):
                self._log_exception(
                    (exc_type, exc_value, exc_tb),
                    background=True
                )
        finally:
            return self._orig_sys_excepthook(exc_type, exc_value, exc_tb)

    @_log_errors
    def _thread_excepthook(self, args):
        """Unhandled exception in thread hook."""
        try:
            if self._should_log(args.exc_value):
                self._log_exception(
                    (args.exc_type, args.exc_value, args.exc_traceback),
                    background=True, thread=args.thread
                )
        finally:
            return self._orig_thread_excepthook(args)

    @_log_errors
    def log_user_input(self, info):
        """Logs the most recent input by way of the 'pre_run_cell' hook."""
        self.line_in_progress = True
        line_num = len(self.ipython_in)
        logger.input('In  [%d]: %s', line_num, info.raw_cell)

    @_log_errors
    def _log_exception(
        self, exc_info, line_input="[n/a]", background=False, thread=None
    ):
        """
        Logs the given exception.

        Parameters
        ----------
        exc_info : (type, value, traceback)
            The exception information.

        line_input: str, optional
            The user input associated with the given line.

        background : bool, optional
            Set if the exception happened in the background as part of an
            exception hook external to IPython.

        thread : threading.Thread, optional
            The associated thread when not the main thread, if available.
        """
        exc_type, exc_value, exc_traceback = exc_info
        try:
            line_num = len(self.ipython_in) - 1
            line_traceback = ''.join(traceback.format_exception(*exc_info))
            thread = f" ({thread})" if thread else ""
            logger.input(
                """\
Exception in IPython session%s. Last input line %s:
%s

Exception details:
%s
""".rstrip(),
                thread,
                line_num,
                _indented(line_input),
                _indented(line_traceback),
            )

            log_exception_to_central_server(
                exc_info,
                stacklevel=2,
                message=f"""\
Line: {line_num}{thread}
Input: {line_input}
Exception: {exc_type.__name__}: {exc_value}
""".rstrip(),
            )
        finally:
            self.prev_err_value = exc_value

    @_log_errors
    def log_exception(self, line_input="[n/a]"):
        """Logs the most recent unhandled exception."""
        # These three ``sys.last_*`` variables are not always defined; they are
        # set when an exception is not handled and the interpreter prints an
        # error message and a stack traceback.

        last_value = getattr(sys, "last_value", None)
        if not last_value or not isinstance(last_value, Exception):
            # No exception to log, or ignore non-exception types
            return

        exc_info = (sys.last_type, last_value, sys.last_traceback)
        self._log_exception(exc_info, line_input=line_input)

    @_log_errors
    def log_output(self, result):
        """Logs the most recent output by way of the 'post_run_cell' hook."""
        self.line_in_progress = False
        if result.result:
            # Convert to string, limit to max tweet length
            line_num = len(self.ipython_in) - 1
            last_out = str(result.result)[:280]
            logger.input('Out [%d]: %s', line_num, last_out)

        if result.error_in_exec is not None:
            self.log_exception(result.info.raw_cell)


def load_ipython_extension(ipython):
    """
    Initialize the `IPythonLogger`.

    This involves adding the ``INPUT`` log level and registering
    `IPythonLogger.log` to run on the ``post-execute`` event.

    Parameters
    ----------
    ip: ``ipython`` ``Shell``
        The active ``ipython`` ``Shell``, perhaps the one returned by
        ``IPython.get_ipython()``.
    """
    global _ip_logger
    _ip_logger = IPythonLogger(ipython)



import logging

from hutch_python.user_load import get_user_objs

logger = logging.getLogger(__name__)


def test_user_load():
    logger.debug('test_user_load')
    info = ['sample_module_1', 'sample_module_2.py']
    objs = get_user_objs(info, ask_on_failure=False)
    assert objs['hey'] == '4horses'
    assert objs['milk'] == 'cows'
    assert objs['some_int'] == 5
    assert objs['just_this'] == 5.0
    assert 'not_this' not in objs



import logging
import os.path
import tempfile

import simplejson

from hutch_python.happi import get_happi_objs, get_lightpath

from .conftest import beamlines, requires_lightpath

logger = logging.getLogger(__name__)


@requires_lightpath
def test_happi_objs():
    logger.debug("test_happi_objs")
    db = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                      'happi_db.json')
    # Only select active objects
    objs = get_happi_objs(db, 'tst')
    assert len(objs) == 2
    assert all([obj.active for obj in objs.values()])
    # Check that we can get upstream devices
    beamlines['RBD'] = {'TST': {}}
    objs = get_happi_objs(db, 'rbd')
    assert len(objs) == 2
    assert all([obj.active for obj in objs.values()])
    # Make sure we can handle an empty JSON file
    with tempfile.NamedTemporaryFile('w+') as tmp:
        simplejson.dump(dict(), tmp)
        tmp.seek(0)
        assert get_happi_objs(tmp.name, 'tst') == {}


@requires_lightpath
def test_get_lightpath():
    logger.debug("test_get_lightpath")
    db = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                      'happi_db.json')
    obj = get_lightpath(db, 'tst')
    # Check that we created a valid BeamPath with no inactive objects
    assert obj.name == 'TST'
    assert len(obj.devices) == 2



import logging

import happi
from happi.loader import load_devices

try:
    import lightpath
    from lightpath.config import beamlines
except ImportError:
    lightpath = None
    beamlines = None

logger = logging.getLogger(__name__)


def get_happi_objs(db, hutch):
    """
    Get the relevant items for ``hutch`` from ``db``.

    This depends on a JSON ``happi`` database stored somewhere in the file
    system and handles setting up the ``happi.Client`` and querying the data
    base for items.

    Parameters
    ----------
    db: ``str``
        Path to database

    hutch: ``str``
        Name of hutch

    Returns
    -------
    objs: ``dict``
        A mapping from item name to item
    """
    # Load the happi Client
    client = happi.Client(path=db)
    containers = list()
    if beamlines is None:
        beamline_conf = []
    else:
        # Find upstream items based on lightpath configuration
        beamline_conf = beamlines.get(hutch.upper())
    # Something strange is happening if there are no upstream items
    if not beamline_conf:
        logger.warning("Unable to find lightpath for %s",
                       hutch.upper())
        beamline_conf = {}
    # Add the complete hutch beamline
    beamline_conf[hutch.upper()] = {}
    # Base beamline
    for beamline, conf in beamline_conf.items():
        # Assume we want hutch items that are active
        reqs = dict(beamline=beamline, active=True)
        results = client.search(**reqs)
        blc = [res.item for res in results]
        # Add the beamline containers to the complete list
        if blc:
            containers.extend(blc)
        else:
            logger.warning("No items found in database for %s",
                           beamline.upper())
    # Instantiate the devices needed
    dev_namespace = load_devices(*containers, pprint=False)
    return dev_namespace.__dict__


def get_lightpath(db, hutch):
    """
    Create a lightpath from relevant ``happi`` objects.

    Parameters
    ----------
    db: ``str``
        Path to database

    hutch: ``str``
        Name of hutch

    Returns
    -------
    path: ``lightpath.BeamPath``
        Object that provides a convenient way to visualize all the devices
        that may block the beam on the way to the interaction point.
    """
    if None in (lightpath, beamlines):
        logger.warning('Lightpath module is not available.')
        return None
    # Load the happi Client
    client = happi.Client(path=db)
    # Allow the lightpath module to create a path
    lc = lightpath.LightController(client, endstations=[hutch.upper()])
    # Return the BeamPath object created by the LightController
    return lc.beamlines[hutch.upper()]



from importlib import import_module

from . import utils


def collect_plans(modules):
    """
    Take all the plans in ``modules`` and collect them into a namespace.

    Arguments
    ---------
    modules: ``list of str``
        The modules to extract plans from.
    """
    plans = {}
    for module_name in modules:
        module = import_module(module_name)
        for name, obj in module.__dict__.items():
            try:
                # Only include things that are natively from this module
                from_module = obj.__module__ == module_name
                # Only include callables
                is_callable = callable(obj)
                # Skip hidden items
                is_hidden = len(name) == 0 or name[0] == '_'

                if from_module and is_callable and not is_hidden:
                    plans[name] = obj
            except AttributeError:
                # obj did not have __module__, probably a builtin
                pass
    return utils.HelpfulNamespace(**plans)


plans = collect_plans(['bluesky.plans',
                       'nabs.plans'])
plan_stubs = collect_plans(['bluesky.plan_stubs',
                            'nabs.plan_stubs'])
preprocessors = collect_plans(['bluesky.preprocessors',
                               'nabs.preprocessors'])



"""
This module defines namespace mappings for debug utilities.
"""
from .log_setup import (debug_context, debug_mode, debug_wrapper,
                        set_console_level)


debug_tools = {
    'debug_console_level': set_console_level,
    'debug_mode': debug_mode,
    'debug_context': debug_context,
    'debug_wrapper': debug_wrapper,
    }


def load_debug(cache):
    cache(**debug_tools)



"""
This module defines options and settings the user may opt to change.
"""
from pcdsdevices.interface import set_engineering_mode


def set_default_options():
    set_engineering_mode(False)


def load_options(cache):
    cache(set_engineering_mode=set_engineering_mode)
    cache.doc(set_engineering_mode='Tab interface verbosity settings.')
    set_default_options()



"""
This module defines the command-line interface arguments for the
``hutch-python`` script. It also provides utilities that are only used at
startup.
"""
import argparse
import logging
import os
from pathlib import Path

import IPython
import matplotlib
from cookiecutter.main import cookiecutter
from IPython import start_ipython
from traitlets.config import Config

from .constants import CONDA_BASE, DIR_MODULE
from .load_conf import load
from .log_setup import configure_log_directory, debug_mode, setup_logging

logger = logging.getLogger(__name__)
opts_cache = {}

# Define the parser
parser = argparse.ArgumentParser(prog='hutch-python',
                                 description='Launch LCLS Hutch Python')
parser.add_argument('--cfg', required=False, default=None,
                    help='Configuration yaml file')
parser.add_argument('--exp', required=False, default=None,
                    help='Experiment number override')
parser.add_argument('--debug', action='store_true', default=False,
                    help='Start in debug mode')
parser.add_argument('--sim', action='store_true', default=False,
                    help='Run with simulated DAQ (lcls1 only)')
parser.add_argument('--create', action='store', default=False,
                    help='Create a new hutch deployment')
parser.add_argument('script', nargs='?',
                    help='Run a script instead of running interactively')

# Append to module docs
__doc__ += '\n::\n\n    ' + parser.format_help().replace('\n', '\n    ')


def configure_tab_completion(ipy_config):
    """
    Disable Jedi and tweak IPython tab completion.

    Parameters
    ----------
    ipy_config : traitlets.config.Config
        IPython configuration.
    """
    # Old API for disabling Jedi. Keep in just in case API changes back.
    ipy_config.InteractiveShellApp.Completer.use_jedi = False
    # New API for disabling Jedi (two access points documented, use both)
    ipy_config.Completer.use_jedi = False
    ipy_config.IPCompleter.use_jedi = False
    try:
        # Monkeypatch IPython completion - we need it to respect __dir__
        # when Jedi is disabled.
        # Details: https://github.com/pcdshub/pcdsdevices/issues/709
        # First, access it to see that the internals have not changed:
        IPython.core.completer.dir2
    except AttributeError:
        logger.debug('Looks like the IPython API changed!')
    else:
        # Then monkeypatch it in:
        IPython.core.completer.dir2 = dir


def configure_ipython_session():
    """
    Configure a new IPython session.

    Returns
    -------
    ipy_config : traitlets.config.Config
        IPython configuration.
    """
    ipy_config = Config()
    # Important Utilities
    ipy_config.InteractiveShellApp.extensions = [
        'hutch_python.ipython_log',
        'hutch_python.bug'
    ]
    # Matplotlib setup for ipython (automatically do %matplotlib)
    backend = matplotlib.get_backend().replace('Agg', '').lower()
    ipy_config.InteractiveShellApp.matplotlib = backend
    if backend == 'agg':
        logger.warning('No matplotlib rendering available. '
                       'Methods that create plots will not '
                       'function properly.')

    configure_tab_completion(ipy_config)
    return ipy_config


def main():
    """
    Do the full hutch-python launch sequence.

    Parses the user's cli arguments and distributes them as needed to the
    setup functions.
    """
    # Parse the user's arguments
    args = parser.parse_args()

    # Set up logging first
    if args.cfg is None:
        log_dir = None
    else:
        log_dir = os.path.join(os.path.dirname(args.cfg), 'logs')

    configure_log_directory(log_dir)
    setup_logging()

    # Debug mode next
    if args.debug:
        debug_mode(True)

    # Do the first log message, now that logging is ready
    logger.debug('cli starting with args %s', args)

    # Options that mean skipping the python environment
    if args.create:
        hutch = args.create
        envs_dir = CONDA_BASE / 'envs'
        if envs_dir.exists():
            # Pick most recent pcds release in our common env
            base = str(CONDA_BASE)
            path_obj = sorted(envs_dir.glob('pcds-*'))[-1]
            env = path_obj.name
        else:
            # Fallback: pick current env
            base = str(Path(os.environ['CONDA_EXE']).parent.parent)
            env = os.environ['CONDA_DEFAULT_ENV']
        logger.info(('Creating hutch-python dir for hutch %s using'
                     ' base=%s env=%s'), hutch, base, env)
        cookiecutter(str(DIR_MODULE / 'cookiecutter'), no_input=True,
                     extra_context=dict(base=base, env=env, hutch=hutch))
        return

    # Save whether we are an interactive session or a script session
    opts_cache['script'] = args.script

    # Load objects based on the configuration file
    objs = load(cfg=args.cfg, args=args)

    script = opts_cache.get('script')
    if script is None:
        # Finally start the interactive session
        start_ipython(argv=['--quick'], user_ns=objs,
                      config=configure_ipython_session())
    else:
        # Instead of setting up ipython, run the script with objs
        with open(script) as fn:
            code = compile(fn.read(), script, 'exec')
            exec(code, objs, objs)



import logging
import os.path
from configparser import ConfigParser, NoOptionError

import happi
from happi.loader import load_devices

from .utils import safe_load

try:
    from happi.backends.qs_db import QSBackend
except ImportError:
    # Optional because not available on windows
    QSBackend = None

logger = logging.getLogger(__name__)


def get_qs_objs(expname):
    """
    Gather user objects from the experiment questionnaire.

    Parameters
    ----------
    expname: ``str``
        The experiment's name from the elog

    Returns
    -------
    objs: ``dict``
        Mapping from questionnaire ``python name`` to loaded object.
    """
    logger.debug('get_qs_client(%s)', expname)
    with safe_load('questionnaire'):
        expname = expname.lower()
        qs_client = get_qs_client(expname)
        # Create namespace
        if not qs_client.all_items:
            logger.warning("No items found in PCDS Questionnaire for %s",
                           expname)
            return dict()
        dev_namespace = load_devices(*qs_client.all_items, pprint=False)
        return dev_namespace.__dict__
    return {}


def get_qs_client(expname):
    """
    Create a `happi.Client` from the experiment questionnaire.

    Connects to the questionnaire webservice via the ``happi`` ``QSBackend``
    using ``psdm_qs_cli`` to collect well-defined devices.

    There are two possible methods of authentication to the
    ``QuestionnaireClient``, ``Kerberos`` and ``WS-Auth``. The first is simpler
    but is not possible for all users, we therefore search for a configuration
    file named ``web.cfg``, either hidden in the current directory or the users
    home directory. This should contain the username and password needed to
    authenticate into the ``QuestionnaireClient``. The format of this
    configuration file is the standard ``.ini`` structure and should define the
    username and password like:

    .. code:: ini

        [DEFAULT]
        user = MY_USERNAME
        pw = MY_PASSWORD

    Parameters
    ----------
    expname: ``str``
        The experiment's name from the elog

    Returns
    -------
    qs_client: `happi.Client`
        Mapping from questionnaire ``python name`` to loaded object.
    """
    if QSBackend is None:
        raise RuntimeError('psdm_qs_cli library unavailable')
    # Determine which method of authentication we are going to use.
    # Search for a configuration file, either in the current directory
    # or hidden in the users home directory. If not found, attempt to
    # launch the client via Kerberos
    cfg = ConfigParser()
    cfgs = cfg.read(['qs.cfg', '.qs.cfg',
                     os.path.expanduser('~/.qs.cfg'),
                     'web.cfg', '.web.cfg',
                     os.path.expanduser('~/.web.cfg')])
    # Ws-auth
    if cfgs:
        user = cfg.get('DEFAULT', 'user', fallback=None)
        try:
            pw = cfg.get('DEFAULT', 'pw')
        except NoOptionError as exc:
            raise ValueError("Must specify password as 'pw' in "
                             "configuration file") from exc
        qs_client = happi.Client(database=QSBackend(expname,
                                                    use_kerberos=False,
                                                    user=user, pw=pw))
    # Kerberos
    else:
        qs_client = happi.Client(database=QSBackend(expname,
                                                    use_kerberos=True))
    return qs_client



import logging
import os
import sys
from collections import namedtuple
from contextlib import contextmanager
from copy import copy
from logging.handlers import QueueHandler
from pathlib import Path
from queue import Queue

import pytest
import zmq
from epics import PV
from ophyd.areadetector.plugins import PluginBase
from ophyd.device import Component as Cpt
from ophyd.signal import Signal
from pcdsdevices.areadetector.detectors import PCDSAreaDetector

import hutch_python.qs_load
import hutch_python.utils

try:
    from elog import HutchELog
except ImportError:
    HutchELog = None
try:
    import lightpath
    from lightpath import beamlines
except ImportError:
    lightpath = None
    beamlines = None
try:
    from psdaq.control.BlueskyScan import BlueskyScan
except ImportError:
    BlueskyScan = None


# Some re-usable skip decorators
skip_if_win32_generic = pytest.mark.skipif(
    sys.platform == 'win32',
    reason='Does not run on Windows',
    )
skip_if_win32_pcdsdaq = pytest.mark.skipif(
    sys.platform == 'win32',
    reason='Fails on Windows (pcdsdaq)',
)
requires_lightpath = pytest.mark.skipif(
    lightpath is None,
    reason='lightpath module not installed',
)
requires_elog = pytest.mark.skipif(
    HutchELog is None,
    reason='elog module not installed',
)
requires_psdaq = pytest.mark.skipif(
    BlueskyScan is None,
    reason='psdaq.control not installed',
)

# We need to have the tests directory importable to match what we'd have in a
# real hutch-python install
sys.path.insert(0, os.path.dirname(__file__))

TST_CAM_CFG = str(Path(__file__).parent / '{}camviewer.cfg')

for component in PCDSAreaDetector.component_names:
    cpt_class = getattr(PCDSAreaDetector, component).cls
    if issubclass(cpt_class, PluginBase):
        cpt_class.plugin_type = Cpt(Signal, value=cpt_class._plugin_type)

# Stupid patch that somehow makes the test cleanup bug go away
PV.count = property(lambda self: 1)


@contextmanager
def cli_args(args):
    """
    Context manager for running a block of code with a specific set of
    command-line arguments.
    """
    prev_args = sys.argv
    sys.argv = args
    yield
    sys.argv = prev_args


@contextmanager
def restore_logging():
    """
    Context manager for reverting our logging config after testing a function
    that configures the logging.
    """
    prev_handlers = copy(logging.root.handlers)
    yield
    logging.root.handlers = prev_handlers


@pytest.fixture(scope='function')
def log_queue():
    with restore_logging():
        my_queue = Queue()
        handler = QueueHandler(my_queue)
        root_logger = logging.getLogger('')
        root_logger.addHandler(handler)
        yield my_queue


Experiment = namedtuple('Experiment', ('run', 'proposal',
                                       'user', 'pw', 'kerberos'))


class QSBackend:
    empty = False

    def __init__(self, expname, use_kerberos=True, user=None, pw=None):
        self.expname = expname
        self.user = user
        self.pw = pw
        self.kerberos = use_kerberos

        if 'bad' in expname:
            raise RuntimeError('bad expname')

    def find(self, to_match):
        device = {
            '_id': 'TST:USR:MMN:01',
            'beamline': 'TST',
            'device_class': 'hutch_python.tests.conftest.Experiment',
            'location': 'Hutch-main experimental',
            'args': ['{{run}}', '{{proposal}}',
                     '{{user}}', '{{pw}}', '{{kerberos}}'],
            'kwargs': {},
            'name': 'inj_x',
            'prefix': 'TST:USR:MMN:01',
            'purpose': 'Injector X',
            'type': 'pcdsdevices.happi.containers.LCLSItem',
            'run': self.expname[-2:],
            'user': self.user,
            'pw': self.pw,
            'kerberos': self.kerberos,
            'proposal': self.expname[3:-2].upper()}
        if self.empty:
            return
        else:
            yield device
            return

    # Dummy methods to make this "look like a database"
    def all_devices(self, *args, **kwargs):
        pass

    def all_items(self, *args, **kwargs):
        pass

    def delete(self, *args, **kwargs):
        pass

    def save(self, *args, **kwargs):
        pass


@pytest.fixture(scope='function')
def fake_qsbackend(monkeypatch):
    monkeypatch.setattr(hutch_python.qs_load, "QSBackend", QSBackend)
    QSBackend.empty = False
    return QSBackend


cfg = """\
[DEFAULT]
user=user
pw=pw
[GITHUB]
user=github_user
pw=github_pw
proxy=http://proxyhost:11111
"""


@pytest.fixture(scope='function')
def temporary_config():
    # Write to our configuration
    with open('web.cfg', '+w') as f:
        f.write(cfg)
    # Allow the test to run
    yield
    # Remove the file
    os.remove('web.cfg')


@pytest.fixture(scope='function')
def fake_curexp_script():
    old_script = hutch_python.utils.CUR_EXP_SCRIPT
    hutch_python.utils.CUR_EXP_SCRIPT = 'echo {}lr1215'
    yield
    hutch_python.utils.CUR_EXP_SCRIPT = old_script


if HutchELog is None:
    ELog = None
else:
    class ELog(HutchELog):
        """Pseudo ELog"""
        def __init__(self, instrument, station=None, user=None, pw=None):
            self.instrument = instrument
            self.station = station
            self.user = user
            self.pw = pw


class DummyZMQContext:
    def __init__(self, *args, **kwargs):
        pass

    def socket(self, *args, **kwargs):
        return DummyZMQSocket()


class DummyZMQSocket:
    def __init__(self, *args, **kwargs):
        pass

    def connect(self, *args, **kwargs):
        pass

    def setsockopt(self, *args, **kwargs):
        pass

    def close(self, *args, **kwargs):
        pass

    def bind(self, *args, **kwargs):
        pass

    def send_json(self, *args, **kwargs):
        pass

    def recv_json(self, *args, **kwargs):
        return {}


@pytest.fixture(scope='function')
def dummy_zmq_lcls2(monkeypatch):
    # monkeypatch such that the lcls2 daq doesn't actually connect to things
    monkeypatch.setattr(zmq, 'Context', DummyZMQContext)



import logging

from ophyd.areadetector.base import EpicsSignalWithRBV
from ophyd.sim import make_fake_device, fake_device_cache, FakeEpicsSignal
import pytest

from pcdsdevices.areadetector.detectors import PCDSAreaDetector

import hutch_python.cam_load as cam_load
from hutch_python.cam_load import (read_camviewer_cfg, interpret_lines,
                                   build_and_log, build_cam,
                                   UnsupportedConfig, MalformedConfig)

from .conftest import TST_CAM_CFG

logger = logging.getLogger(__name__)
CFG = TST_CAM_CFG.format('')

fake_device_cache[EpicsSignalWithRBV] = FakeEpicsSignal
FakeDet = make_fake_device(PCDSAreaDetector)
cam_load.PCDSAreaDetector = FakeDet


def test_build_cam():
    logger.debug('test_build_cam')
    # Basic functionality test
    obj = build_cam('GE:16', 'PREFIX:IMAGE2', None, 'my_cam')
    assert isinstance(obj, PCDSAreaDetector)


def test_build_cam_errors():
    logger.debug('test_build_cam_errors')
    # Cover bad configs
    with pytest.raises(UnsupportedConfig):
        build_cam('asfd', '', '', '')

    with pytest.raises(MalformedConfig):
        build_cam('GE', '', '', '')


def test_read_camviewer_cfg():
    logger.debug('test_read_camviewer_cfg')
    # Basic functionality test
    objs = read_camviewer_cfg(CFG)
    assert isinstance(objs['my_cam'], PCDSAreaDetector)
    assert len(objs) == 1


def test_include():
    logger.debug('test_include')
    info = interpret_lines(['include ' + CFG,
                            'include'])
    assert len(info) == 4


def test_bad_object():
    logger.debug('test_bad_object')
    # Shouldn't get a full raise out of build_and_log when obj is bad
    build_and_log(['GE', logger, None, 'bad'])



import logging

import happi

from hutch_python.qs_load import get_qs_objs

logger = logging.getLogger(__name__)


def clear_happi_cache():
    happi.loader.cache = {}


def test_qs_load(fake_qsbackend):
    logger.debug('test_qs_load')
    clear_happi_cache()
    objs = get_qs_objs('tstlr1215')
    assert objs['inj_x'].run == '15'
    assert objs['inj_x'].proposal == 'LR12'
    assert objs['inj_x'].kerberos == 'True'
    # Check that we can handle an empty Questionnaire
    fake_qsbackend.empty = True
    assert get_qs_objs('tstlr1215') == dict()
    fake_qsbackend.empty = False


def test_ws_auth_conf(temporary_config, fake_qsbackend):
    logger.debug('test_ws_auth_conf')
    clear_happi_cache()
    objs = get_qs_objs('tstlr1215')
    assert objs['inj_x'].kerberos == 'False'
    assert objs['inj_x'].user == 'user'
    assert objs['inj_x'].pw == 'pw'



from pcdsdevices.beam_stats import LCLS, BeamStats


def global_devices():
    """
    Instantiates information devices shared across LCLS.
    """
    return dict(
        beam_stats=BeamStats(),
        lcls=LCLS()
        )


global_device_docs = {
    'beam_stats': 'Summary of the most important beam statistics.',
    'lcls': 'Collection of all upstream x-ray diagnostics.',
    }



# If no environment variable set, try qt5 (best for hutch-python)
import os
os.environ.setdefault('MPLBACKEND', 'Qt5Agg')
# Matplotlib will choose an appropriate backend based on the environment
# variable settings and its ability to communicate with an X server.
# Our preference, as above, is to use the Qt5 backend when available.
import matplotlib.pyplot  # noqa: F401 E402



from pcdsdevices.sim import FastMotor, SlowMotor

from .utils import HelpfulNamespace


def get_hw():
    """
    Create and return a standard namespace of simulated hardware.
    """
    ns = HelpfulNamespace(
        fast_motor1=FastMotor(name='fast_motor1'),
        fast_motor2=FastMotor(name='fast_motor2'),
        fast_motor3=FastMotor(name='fast_motor3'),
        slow_motor1=SlowMotor(name='slow_motor1'),
        slow_motor2=SlowMotor(name='slow_motor2'),
        slow_motor3=SlowMotor(name='slow_motor3'))
    return ns



"""
Load Detector objects based on a camviewer config file
"""
import asyncio
import logging
from functools import partial
from multiprocessing.pool import ThreadPool
from multiprocessing import cpu_count

from .constants import SUCCESS_LEVEL

from pcdsdevices.areadetector.detectors import PCDSAreaDetector

logger = logging.getLogger(__name__)
logger.success = partial(logger.log, SUCCESS_LEVEL)

main_event_loop = None


def read_camviewer_cfg(filename):
    """
    Read camviewer.cfg file and create detector objects.

    Parameters
    ----------
    filename: ``str``
        Full path to the camviewer.cfg file.

    Returns
    -------
    objs: ``{str: PCDSAreaDetector}``
        Each detector object, indexed by name.
    """
    info = interpret_cfg(filename)
    return load_cams(info)


def interpret_cfg(filename, pvnames=None):
    """
    Read camviewer.cfg file and interpret lines

    Parameters
    ----------
    filename: ``str``
        Full path to the camviewer.cfg file.

    Returns
    -------
    info: ``list of str``
        Valid inputs for build_cam
    """
    with open(filename, 'r') as f:
        lines = f.readlines()

    if pvnames is None:
        pvnames = set()
    return interpret_lines(lines, pvnames=pvnames)


def interpret_lines(lines, pvnames=None):
    """
    Create info list from string lines of a camviewer.cfg file.

    Parameters
    ----------
    lines: ``list of str``
        The python strings for each line of the config file

    pvnames: ``set``, optional
        For internal use only, to keep track of which PVs we're already
        planning to load so we don't make duplicate objects.

    Returns
    -------
    info: ``list of list str``
        Valid inputs for build_cam
    """
    info = []
    if pvnames is None:
        pvnames = set()

    for line in lines:
        line = line.strip()
        if not line or line[0] == '#':
            continue

        logger.debug(line)
        parts = line.split(',')
        parts = [p.strip() for p in parts]

        if parts[0].startswith('include'):
            try:
                inc = interpret_cfg(parts[0].split(' ')[1], pvnames=pvnames)
                info.extend(inc)
            except IndexError:
                err = 'Malformed include line "%s" in camviewer cfg, skipping.'
                logger.error(err, line)
                logger.debug(err, line, exc_info=True)
        elif len(parts) > 1:
            logger.debug(parts)
            pvname = get_det_prefix(parts[1])
            if pvname not in pvnames:
                logger.debug(pvname)
                pvnames.add(pvname)
                info.append(parts)

    return info


def load_cams(info):
    """
    Create detector object from string lines of a camviewer.cfg file.

    Parameters
    ----------
    info: ``list of list of str``
        Valid inputs for build_cam

    Returns
    -------
    objs: ``{str: PCDSAreaDetector}``
        Each detector object, indexed by name.
    """
    # Any cam that needs the event loop has to have one set in the thread
    global main_thread_loop
    main_thread_loop = asyncio.get_event_loop()

    objs = {}
    logger.debug(info)
    pool = ThreadPool(cpu_count()-1)
    obj_list = pool.map(build_and_log, info)
    for obj in obj_list:
        if obj is not None:
            objs[obj.name] = obj
    return objs


def build_and_log(info_part):
    """
    Wrapper to run build_cam in a threadpool and handle errors

    Parameters
    ----------
    info_part: ``list of str``
        Valid input for build_cam

    Returns
    -------
    obj: ``PCDSAreaDetector``
        The loaded detector, or None.
    """
    # We sync with the main thread's loop so that they work as expected later
    asyncio.set_event_loop(main_thread_loop)
    try:
        obj = build_cam(*info_part)
        logger.success('Loaded %s', obj.name)
        return obj
    except UnsupportedConfig as exc:
        logger.debug('Skip non area detector cam %s',
                     exc.name, exc_info=True)
    except MalformedConfig:
        err = 'Skip malformed config %s'
        logger.error(err, info_part)
        logger.debug(err, info_part, exc_info=True)
    except TypeError:
        err = 'Malformed config %s in camviewer cfg, skipping.'
        logger.error(err, info_part)
        logger.debug(err, info_part, exc_info=True)
    except Exception:
        if len(info_part) >= 4:
            name = info_part[3]
        else:
            name = info_part
        err = ('Error loading %s in camviewer cfg. IOC '
               'probably down, skipping.')
        logger.error(err, name)
        logger.debug(err, info_part, exc_info=True)


def build_cam(cam_type, pv_info, evr, name, *args):
    """
    Create a single detector object, using camviewer.cfg information.

    Paramaters
    ----------
    cam_type: ``str``
        The first element in the camviewer config line, determines whether or
        not we have a detector class for the camera.

    pv_info: ``str``
        Second element in the line, determines which PVs to use.

    evr: ``str``
        Third element in the line, determines the EVR PVs. This is currently
        not used.

    name: ``str``
        Fourth element in th eline, determines the object's name.

    *args:
        Optional arguments that are unused.

    Returns
    -------
    cam: ``PCDSAreaDetector``
    """
    if not cam_type.startswith('GE'):
        raise UnsupportedConfig('Only cam type GE (area detector) supported.',
                                name=name)
    if not pv_info or not name:
        raise MalformedConfig(name=name)
    detector_prefix = get_det_prefix(pv_info)
    name = name.replace(' ', '_').lower()
    return PCDSAreaDetector(detector_prefix, name=name)


def get_det_prefix(pv_info):
    """
    Determines which prefix will be passed into the detector init.

    Parameters
    ----------
    pv_info: ``str``
        The second element in a camview cfg line.

    Returns
    -------
    detector_prefix: ``str``
        The prefix to use in the detector init.
    """
    pv_info = pv_info.split(';')
    try:
        detector_prefix = pv_info[1]
    except IndexError:
        # Not provided in config, guess from image base
        detector_prefix = ':'.join(pv_info[0].split(':')[:-1])
    return detector_prefix + ':'


class UnsupportedConfig(Exception):
    def __init__(self, message=None, name=None):
        self.message = message or 'Unsupported config'
        self.name = name


class MalformedConfig(Exception):
    def __init__(self, message=None, name=None):
        self.message = message or 'Malformed config'
        self.name = name



from importlib import import_module

from .utils import HelpfulNamespace


def collect_functions(modules):
    """
    Take all the callables in ``modules`` and collect them into a namespace.

    Arguments
    ---------
    modules: ``list of str``
        The modules to extract functions from
    """
    functions = {}
    for module_name in modules:
        module = import_module(module_name)
        for name, obj in module.__dict__.items():
            try:
                # Only include things that are natively from this module
                from_module = obj.__module__ == module_name
                # Only include callables
                is_callable = callable(obj)
                # Skip hidden items
                is_hidden = len(name) == 0 or name[0] == '_'

                if from_module and is_callable and not is_hidden:
                    functions[name] = obj
            except AttributeError:
                # obj did not have __module__, probably a builtin
                pass
    return HelpfulNamespace(**functions)


calc_namespace = HelpfulNamespace(
    be_lens=collect_functions(['pcdscalc.be_lens_calcs']),
    common=collect_functions(['pcdscalc.common']),
    diffraction=collect_functions(['pcdscalc.diffraction']),
    )



import logging
import os
import pathlib
import tempfile

import simplejson
from requests import Response

import hutch_python.bug
from hutch_python.bug import (get_current_environment, get_text_from_editor,
                              report_bug)

from .conftest import skip_if_win32_generic

logger = logging.getLogger(__name__)


# Mock user input function that just returns the prompt
def user_input(prompt):
    return prompt


# Mock requests.Session to avoid actual GitHub posts
class FakeSession:
    handle = None
    past_auth = dict()
    proxies = dict()

    def __init__(self):
        pass

    @property
    def auth(self):
        return self.past_auth

    @auth.setter
    def auth(self, auth_info):
        logger.debug('Updating authorization %s', auth_info)
        self.past_auth.update(dict([auth_info]))

    def post(self, url, json_dump):
        # Dump the JSON to handle
        if self.handle:
            logger.debug('Dumping JSON to %s', self.handle.name)
            self.handle.write(json_dump)
        # Create a fake response
        r = Response()
        r.status_code = 201
        return r


def test_get_current_environment():
    logger.debug('test_get_current_environment')
    # Development packages
    fk_pkgs = ['one', 'two', 'three']
    os.environ['CONDA_ENVNAME'] = 'test-environment'
    with tempfile.TemporaryDirectory() as tmp:
        logger.debug('Creating temp directory %s', tmp)
        # Set a PythonPath
        os.environ['PYTHONPATH'] = tmp
        # Create fake packages
        for pkg in fk_pkgs:
            pathlib.Path(os.path.join(tmp, pkg)).touch()
        env, pkgs = get_current_environment()
        assert all([pkg in pkgs for pkg in fk_pkgs])
        assert env == 'test-environment'


def test_bug_report(monkeypatch, temporary_config):
    logger.debug('test_bug_report')
    # Patch in our fake user input function
    monkeypatch.setattr(hutch_python.bug, 'request_input', user_input)
    # Patch in fake requests.Session
    monkeypatch.setattr(hutch_python.bug.requests, 'Session', FakeSession)
    # Create a fake issue and requests.Session
    with tempfile.NamedTemporaryFile('w+', delete=False) as tmp:
        # Write to our fake file instead of GitHub
        FakeSession.handle = tmp
        # Post report
        report_bug(captured_output='A printed message',
                   description='A description of the bug')
    bug = simplejson.load(open(tmp.name, 'r'))
    # Check GitHub authentication
    assert FakeSession.past_auth == {'github_user': 'github_pw'}
    # Test proxy host information
    assert FakeSession.proxies == {'https': 'http://proxyhost:11111'}
    # See that we catch the fake environment
    assert bug['title'].startswith('Please')


@skip_if_win32_generic
def test_get_text_from_editor(monkeypatch):
    logger.debug("test_get_text_from_editor")

    # Mock write function
    def write_text(info):
        open(info[1], 'w+').write(info[0])

    # Patch in our fake write_text
    monkeypatch.setattr(hutch_python.bug.subprocess, 'call', write_text)
    os.environ['EDITOR'] = 'vim'
    assert get_text_from_editor() == 'vim'



import logging
from pathlib import Path
from subprocess import STDOUT, CalledProcessError, check_output

import pytest

from .conftest import skip_if_win32_generic

logger = logging.getLogger(__name__)
tstpython = Path(__file__).parent / 'tstpython'


@skip_if_win32_generic
def test_tstpython_scripts():
    logger.debug('test_tstpython_scripts')

    good_text = check_output([tstpython, 'script.py'])
    assert b'script ran' in good_text

    try:
        bad_text = check_output([tstpython, 'bad_script.py'], stderr=STDOUT)
    except CalledProcessError as err:
        bad_text = err.output
    assert b'Traceback' in bad_text
    assert b'ZeroDivisionError' in bad_text


try:
    import colorama  # noqa
    has_colorama = True
except ImportError:
    has_colorama = False


# See bluesky.log, ophyd.log for places where colorama is initialized
# This is supposed to be safe but apparently colorama is buggy
@pytest.mark.skipif(has_colorama,
                    reason=('IPython breaks in a pseudo-tty if any package '
                            'initializes colorama, ruining this test.'))
def test_tstpython_ipython():
    logger.debug('test_tstpython_ipython')

    # This should show the banner, check if the name unique_device exists, and
    # then exit. There should be no NameError.
    ipy_text = check_output([tstpython], universal_newlines=True,
                            input='unique_device\n')
    assert 'IPython' in ipy_text
    assert 'NameError' not in ipy_text



from pathlib import Path

CAMVIEWER_CFG = '/reg/g/pcds/pyps/config/{}/camviewer.cfg'

CONDA_BASE = Path('/reg/g/pcds/pyps/conda/py36')

ENG_TOOLS = '/reg/g/pcds/engineering_tools/'
CUR_EXP_SCRIPT = ENG_TOOLS + '{0}/scripts/get_curr_exp -H {0}'

CLASS_SEARCH_PATH = ['pcdsdevices.device_types']

EPICS_ARCH_FILE_PATH = '/cds/group/pcds/dist/pds/{}/misc/'

DIR_MODULE = Path(__file__).resolve().parent

FILE_YAML = DIR_MODULE / 'logging.yml'

HUTCH_COLORS = dict(
    amo='38;5;27',
    sxr='38;5;250',
    xpp='38;5;40',
    xcs='38;5;93',
    mfx='38;5;202',
    cxi='38;5;196',
    mec='38;5;214')

INPUT_LEVEL = 5

SUCCESS_LEVEL = 35

VALID_KEYS = ('hutch', 'db', 'load', 'experiment', 'daq_platform')
NO_LOG_EXCEPTIONS = (KeyboardInterrupt, SystemExit)
LOG_DOMAINS = {".pcdsn", ".slac.stanford.edu"}



import dataclasses
import logging
import sys
from queue import Empty, Queue
from typing import Any, Dict, List, Optional

import pytest

from hutch_python.ipython_log import IPythonLogger
from hutch_python.ipython_log import logger as ipython_logger

logger = logging.getLogger(__name__)


@pytest.fixture(scope='function')
def fake_ipython():
    # Clear the sys errors, potentially from previous tests
    try:
        del sys.last_type
        del sys.last_value
        del sys.last_traceback
    except AttributeError:
        pass
    return FakeIPython()


class FakeIPythonEvents:
    def __init__(self):
        self.callbacks = {}

    def register(self, name, callback):
        self.callbacks.setdefault(name, []).append(callback)
        logger.info("Registered callback %s %s", name, callback)


@dataclasses.dataclass
class FakeExecutionInfo:
    """A set of information which comes out of pre_run_cell."""
    # raw_cell is the user-executed code, and all we really care about
    raw_cell: str
    store_history: bool = False
    silent: bool = False
    shell_futures: bool = True


@dataclasses.dataclass
class FakeExecutionResult:
    """A result which comes out of post_run_cell."""
    info: FakeExecutionInfo
    result: Any
    error_in_exec: bool
    execution_count: int = 0
    error_before_exec: bool = False


class FakeIPython:
    """A fake replacement of IPython's ``TerminalInteractiveShell``."""
    user_ns: Dict[str, Any]
    events: FakeIPythonEvents

    def __init__(self):
        self.user_ns = dict(In=[""])
        self.events = FakeIPythonEvents()

    def add_line(self, in_line, out_line=None, is_error=False):
        line_number = len(self.user_ns["In"])
        info = FakeExecutionInfo(raw_cell=in_line)
        result = FakeExecutionResult(
            info=info, result=out_line, error_in_exec=is_error
        )
        logger.info(f"Adding line: {line_number} {in_line} -> {out_line}")
        for cb in self.events.callbacks.get("pre_run_cell", []):
            logger.info(f"Calling {cb.__name__} with args {info}")
            cb(info)
        self.user_ns["In"].append(in_line)
        for cb in self.events.callbacks.get("post_run_cell", []):
            logger.info(f"Calling {cb.__name__} with args {result}")
            cb(result)


@pytest.fixture(scope='function')
def ipylog(log_queue, fake_ipython) -> IPythonLogger:
    """IPythonLogger instance, with a promise of an empty initial queue."""
    ipylog = IPythonLogger(fake_ipython)
    while not log_queue.empty():
        log_queue.get(block=False)
    yield ipylog


def next_pertinent_log_record(
    queue: Queue,
    pertinent_loggers: Optional[List[str]] = None
) -> logging.LogRecord:
    """
    Get the next relevant log record from the queue.

    Parameters
    ----------
    queue : queue.Queue
        The queue to search.

    pertinent_loggers : list of str, optional
        List of logger names that are relevant.
        Defaults to what is being tested in this module - the IPythonLogger
        logger instance.

    Returns
    -------
    record : logging.LogRecord
        The matching record.

    Raises
    ------
    Empty
        If the queue is empty and no relevant messages are available.
    """
    pertinent_loggers = pertinent_loggers or {
        "hutch_python.ipython_log",
    }
    while not queue.empty():
        item = queue.get(block=False)
        if item.name in pertinent_loggers:
            return item

    raise Empty(f"No relevant messages available for {pertinent_loggers}")


def next_pertinent_log_message(
    queue: Queue,
    pertinent_loggers: Optional[List[str]] = None
) -> str:
    """
    Get the next relevant log message from the queue.

    Convenience wrapper around :func:`next_pertinent_log_record`.

    Parameters
    ----------
    queue : queue.Queue
        The queue to search.

    pertinent_loggers : list of str, optional
        List of logger names that are relevant.
        Defaults to what is being tested in this module - the IPythonLogger
        logger instance.

    Returns
    -------
    message : str
        The matching record's message.

    Raises
    ------
    Empty
        If the queue is empty and no relevant messages are available.
    """
    record = next_pertinent_log_record(
        queue, pertinent_loggers=pertinent_loggers
    )
    return record.getMessage()


def assert_no_more_messages(
    queue: Queue,
    pertinent_loggers: Optional[List[str]] = None
) -> None:
    """
    Assert that no more relevant messages are available from a queue.

    Convenience wrapper around :func:`next_pertinent_log_record`.

    Parameters
    ----------
    queue : queue.Queue
        The queue to search.

    pertinent_loggers : list of str, optional
        List of logger names that are relevant.
        Defaults to what is being tested in this module - the IPythonLogger
        logger instance.
    """
    with pytest.raises(Empty):
        next_pertinent_log_record(
            queue, pertinent_loggers=pertinent_loggers
        )


@pytest.mark.timeout(5)
def test_basic(ipylog, log_queue):
    logger.info("Sanity check: ensuring the queue handler works")
    ipython_logger.debug('hello')
    assert 'hello' in next_pertinent_log_record(log_queue).getMessage()
    logger.info("Log message recorded by handler: great!")
    assert_no_more_messages(log_queue)


@pytest.mark.timeout(5)
def test_one_in_no_output(ipylog, log_queue, fake_ipython):
    logger.info("One logged In, no output")
    fake_ipython.add_line('print(5)')
    assert 'In  [1]: print(5)' in next_pertinent_log_message(log_queue)
    assert_no_more_messages(log_queue)


@pytest.mark.timeout(5)
def test_one_in_one_out(ipylog, log_queue, fake_ipython):
    logger.info("One logged In, one Out")
    fake_ipython.add_line('1 + 1', 2)
    assert 'In  [1]: 1 + 1' in next_pertinent_log_message(log_queue)
    assert 'Out [1]: 2' in next_pertinent_log_message(log_queue)
    assert_no_more_messages(log_queue)


@pytest.mark.timeout(5)
def test_zero_division(ipylog, log_queue, fake_ipython):
    logger.info("Set up a ZeroDivisionError for logging")
    try:
        1/0
    except ZeroDivisionError:
        exc_type, exc_value, exc_traceback = sys.exc_info()

    # Force in our exception information, despite the fact that we just handled
    # it above
    sys.last_type = exc_type
    sys.last_value = exc_value
    sys.last_traceback = exc_traceback

    fake_ipython.add_line('1/0', is_error=True)
    assert 'In  [1]: 1/0' in next_pertinent_log_message(log_queue)
    assert "Exception details" in next_pertinent_log_message(log_queue)


@pytest.mark.timeout(5)
def test_forced_failure(ipylog, log_queue, fake_ipython):
    logger.info("OK, now forcing a failure by tweaking internals")
    ipylog.ipython_in = None
    fake_ipython.add_line('something')
    assert 'Logging error' in next_pertinent_log_message(log_queue)



import logging
from logging.handlers import QueueHandler
from pathlib import Path

import ophyd
import pytest
from conftest import restore_logging

from hutch_python import log_setup
from hutch_python.log_setup import (configure_log_directory, debug_context,
                                    debug_mode, debug_wrapper,
                                    get_console_handler, get_console_level,
                                    get_console_level_name, get_debug_handler,
                                    get_session_logfiles, log_objects,
                                    log_objects_off, set_console_level,
                                    setup_logging)

from .conftest import skip_if_win32_generic

logger = logging.getLogger(__name__)


@skip_if_win32_generic
def test_setup_logging():
    logger.debug('test_setup_logging')
    dir_logs = Path(__file__).parent / 'logs'

    with restore_logging():
        setup_logging()

    with restore_logging():
        configure_log_directory(dir_logs)
        setup_logging()

    assert dir_logs.exists()


def test_console_handler(log_queue):
    logger.debug('test_console_handler')

    with pytest.raises(RuntimeError):
        handler = get_console_handler()

    with restore_logging():
        setup_queue_console()
        handler = get_console_handler()
        assert isinstance(handler, QueueHandler)


@skip_if_win32_generic
def test_get_session_logfiles():
    logger.debug('test_get_session_logfiles')
    with restore_logging():
        # Create a parent log file
        configure_log_directory(Path(__file__).parent / 'logs')
        setup_logging()
        debug_handler = get_debug_handler()
        debug_handler.doRollover()
        debug_handler.doRollover()
        assert len(get_session_logfiles()) == 3
        assert all([log.startswith(debug_handler.baseFilename)
                    for log in get_session_logfiles()])


def setup_queue_console():
    root_logger = logging.getLogger('')
    for handler in root_logger.handlers:
        if isinstance(handler, QueueHandler):
            queue_handler = handler
            break
    queue_handler.name = 'console'
    queue_handler.level = 20
    return queue_handler


def clear(queue):
    items = []
    while not queue.empty():
        items.append(queue.get(block=False))
    return items


def assert_is_info(queue):
    clear(queue)
    logger.info('hello')
    logger.debug('goodbye')
    assert 'hello' in queue.get(block=False).getMessage()
    assert queue.empty()


def assert_is_debug(queue):
    clear(queue)
    logger.debug('goodbye')
    assert 'goodbye' in queue.get(block=False).getMessage()


def test_set_console_level(log_queue):
    logger.debug('test_set_console_level')

    setup_queue_console()
    assert_is_info(log_queue)

    # Change console level so we get debug statements
    set_console_level(logging.DEBUG)
    assert get_console_level() == logging.DEBUG
    assert get_console_level_name() == "DEBUG"
    assert_is_debug(log_queue)

    set_console_level("INFO")
    assert get_console_level() == logging.INFO
    assert get_console_level_name() == "INFO"


def test_debug_mode(log_queue):
    logger.debug('test_debug_mode')

    setup_queue_console()
    assert not debug_mode()
    assert_is_info(log_queue)

    debug_mode(debug=True)
    assert debug_mode()
    assert_is_debug(log_queue)

    debug_mode(debug=False)
    assert not debug_mode()
    assert_is_info(log_queue)


def test_debug_context(log_queue):
    logger.debug('test_debug_context')

    setup_queue_console()
    assert_is_info(log_queue)

    with debug_context():
        assert_is_debug(log_queue)

    assert_is_info(log_queue)


def test_debug_wrapper(log_queue):
    logger.debug('test_debug_wrapper')

    setup_queue_console()
    assert_is_info(log_queue)

    debug_wrapper(assert_is_debug, log_queue)

    assert_is_info(log_queue)


def test_log_objects(monkeypatch, log_queue):
    filter = log_setup.ObjectFilter()

    def find_filters():
        yield handler, filter

    handler = setup_queue_console()
    handler.addFilter(filter)

    monkeypatch.setattr(log_setup, "find_root_object_filters",
                        find_filters)

    assert_is_info(log_queue)
    obj = ophyd.ophydobj.OphydObject(name="obj")
    ignored_obj = ophyd.ophydobj.OphydObject(name="ignored_obj")

    # Log messages prior to configuring it
    obj.log.debug("hidden-message-1")
    ignored_obj.log.debug("hidden-message-2")

    # Set the level so we see DEBUG messages in general
    set_console_level("DEBUG")
    # These still should be hidden - we ignore ophyd debug stream in general
    obj.log.info("hidden-message-3")
    ignored_obj.log.info("hidden-message-4")
    obj.log.info("hidden-message-5")
    ignored_obj.log.info("hidden-message-6")

    log_objects(obj, level="DEBUG")
    obj.log.info("shown-message-1")
    obj.log.debug("shown-message-2")

    ignored_obj.log.debug("hidden-message-7")
    ignored_obj.log.info("hidden-message-8")

    log_objects_off()
    obj.log.debug("hidden-message-9")
    obj.log.debug("hidden-message-10")
    messages = [
        msg.getMessage() for msg in clear(log_queue)
    ]
    assert set(messages) == {
        "Recording log messages from obj (level >=DEBUG)",
        "shown-message-1",
        "shown-message-2",
        "No longer recording log messages from obj",
    }



"""
This module provides utilities for grouping objects into namespaces.
"""
from inspect import isfunction
import logging

from ophyd import Device

from .utils import (IterableNamespace, find_class, strip_prefix,
                    extract_objs)

logger = logging.getLogger(__name__)


def class_namespace(cls, scope=None):
    """
    Create a ``namespace`` that contains objects of a specific type.

    Parameters
    ----------
    cls: ``type`` or ``str``

    scope: ``module``, ``namespace``, or ``list`` of these
        Every object attached to the given modules will be considered for the
        `class_namespace`. If ``scope`` is omitted, we'll check all objects
        loaded by ``hutch-python`` and everything in the caller's global frame.
        If anything is an instance of ``ophyd.Device``, we'll also include the
        object's components as part of the scope, using the ``name`` attribute
        to identify them rather than the attribute name on the device. This
        will continue recursively, skipping lazy and dynamic components.

    Returns
    -------
    namespace: `IterableNamespace`
    """
    logger.debug('Create class_namespace cls=%s, scope=%s', cls, scope)
    class_space = IterableNamespace()
    scope_objs = extract_objs(scope=scope, stack_offset=1)

    # Resolve str arguments for cls
    if isinstance(cls, str):
        if cls != 'function':
            try:
                cls = find_class(cls)
            except Exception as exc:
                err = 'Type {} could not be loaded'
                logger.error(err.format(cls))
                logger.debug(exc, exc_info=True)
                return class_space

    # Mapping from Device class to list of lists of attrs that lead us to
    # subdevices of the correct type
    cache = {}

    def inspect_device_cls(device_cls, desired_cls, cache):
        """
        Recursive helper function.

        Gets a list of components of device_class that match desired_cls,
        skipping lazy components. Uses cache to improve speed for hutch-python
        environments with large numbers of objects with the same class.
        """
        try:
            return cache[device_cls]
        except KeyError:
            attrs = []
            if issubclass(device_cls, Device):
                logger.debug('Checking subdevices for class %s', device_cls)
                for cpt_name in device_cls.component_names:
                    cpt = getattr(device_cls, cpt_name)
                    if hasattr(cpt, 'cls') and not cpt.lazy:
                        if issubclass(cpt.cls, desired_cls):
                            attrs.append([cpt_name])
                        subattrs = inspect_device_cls(cpt.cls, desired_cls,
                                                      cache)
                        expand_subattrs = [[cpt_name] + a for a in subattrs]
                        attrs.extend(expand_subattrs)
                if attrs:
                    logger.debug('Class %s has matching subdevices %s',
                                 device_cls, attrs)
                else:
                    logger.debug('Class %s has no matching subdevices',
                                 device_cls)
            cache[device_cls] = attrs
            return attrs

    for name, obj in scope_objs.items():
        # Determine whether or not to include this object
        include = False
        if cls == 'function':
            if isfunction(obj):
                include = True
        elif isinstance(obj, cls):
            include = True

        if include:
            logger.debug('Adding %s to %s namespace', name, cls)
            setattr(class_space, name, obj)

        # Determine whether or not to include any subdevices
        if isinstance(obj, Device):
            subdevice_attrs = inspect_device_cls(obj.__class__, cls, cache)
            for attrs in subdevice_attrs:
                device = obj
                for attr in attrs:
                    device = getattr(device, attr)
                logger.debug('Adding %s to %s namespace',
                             device.name, cls)
                setattr(class_space, device.name, device)

    return class_space


def tree_namespace(scope=None):
    """
    Create a ``namespace`` that accumulates objects and creates a tree.

    This tree is a nested set of `IterableNamespace` objects based on the
    object names as defined in scope. We will split on underscores and use the
    splits to create the tree.

    Parameters
    ----------
    scope: ``module``, ``namespace``, or ``list`` of these
        Every object attached to the given modules will be considered for the
        `tree_namespace`. If ``scope`` is omitted, we'll check all objects
        loaded by ``hutch-python`` and everything in the caller's global frame.

    Returns
    -------
    namespace: `IterableNamespace`
    """
    logger.debug('Create tree_namespace scope=%s', scope)
    tree_space = IterableNamespace()
    scope_objs = extract_objs(scope=scope, stack_offset=1)

    for name, obj in scope_objs.items():
        logger.debug('Add %s to tree namespace', name)
        upper_space = tree_space
        keys = name.split('_')[:-1]

        if keys:
            # Add key to existing namespace branch, create new if needed
            for key in keys:
                name = strip_prefix(name, key)
                # Force lowercase
                key = key.lower()
                if not hasattr(upper_space, key):
                    setattr(upper_space, key, IterableNamespace())
                upper_space = getattr(upper_space, key)
            if hasattr(upper_space, name):
                logger.warning(('Tried to add {} to {}, but something was '
                                'already there. Two devices share the same '
                                'name!'.format(name, upper_space)))
            else:
                setattr(upper_space, name, obj)
    logger.debug('Created tree namespace %s', tree_space)
    return tree_space



import logging
from importlib import import_module

from . import utils

logger = logging.getLogger(__name__)


def get_exp_objs(exp_module, *, ask_on_failure=True):
    """
    Load the correct experiment module.

    This will import User from ``experiments.{exp_module}``
    and create ``User()``, storing it as ``x``.

    Parameters
    ----------
    exp_module: ``str``
        The name of the experiment from the elog, without the hutch. This will
        be the name of the module loaded from the experiments folder.

    ask_on_failure : bool, optional
        If a module fails to load, indicate what happened and ask the user if
        loading should continue.

    Returns
    -------
    user: ``object`` or ``HelpfulNamespace``
        Either the user's class instantiated or a blank namespace for other
        experiment-specific objects to be attached to.
    """
    logger.debug('get_exp_objs(%s)', exp_module)
    module_name = 'experiments.' + exp_module
    with utils.safe_load(exp_module):
        try:
            module = import_module(module_name)
            return module.User()
        except Exception as ex:
            error_ns = utils.HelpfulNamespace()
            error_ns.__doc__ = (
                f"Skipped missing experiment file {exp_module}.py: {ex}"
            )
            import_err = isinstance(ex, ImportError) and module_name in ex.msg
            if import_err or not ask_on_failure:
                logger.info('Skip missing experiment file %s.py', exp_module)
                return error_ns

            utils.maybe_exit(
                logger,
                message=(
                    f'Devices and functions from module {module_name} will NOT'
                    f' be available because it failed to load with:\n\t'
                    f'{ex.__class__.__name__}: {ex}'
                ),
                exception_message=f'Failed to load {module_name}',
            )

    return error_ns



import logging
from types import SimpleNamespace

from hutch_python.exp_load import get_exp_objs

logger = logging.getLogger(__name__)


def test_experiment_objs():
    logger.debug('test_experiment_objs')

    user = get_exp_objs('sample_expname', ask_on_failure=False)
    assert not isinstance(user, SimpleNamespace)

    empty = get_exp_objs('q3qwer', ask_on_failure=False)
    assert isinstance(empty, SimpleNamespace)



"""Module to help create the epicsArch file that will be read by the DAQ."""
import argparse
import logging
import os
import sys

from .qs_load import get_qs_client

from .constants import EPICS_ARCH_FILE_PATH

logger = logging.getLogger(__name__)


def _create_parser():
    """Argument Parser Setup. Define shell commands."""
    parser = argparse.ArgumentParser(description='Create an epicsArch file '
                                                 'from the Quesionnaire')

    parser.add_argument('experiment', help='Experiment name to'
                        ' create the epicsArch file from. E.g.: xpplv6818')

    parser.add_argument('--hutch', action="store",
                        help='Overrides hutch detection from the experiment '
                        'name. E.g.: xpp', default=None)

    parser.add_argument('--path', action="store",
                        help='Directory path where to create the epicsArch '
                        'file. E.g.: /path/to/the/directory/', default=None)

    parser.add_argument('--dry-run', action='store_true', default=False,
                        help='Print to stdout what would be written in the '
                        'archFIle.')
    return parser


def main():
    """Entry point."""
    parser = _create_parser()
    parsed_args = parser.parse_args()
    kwargs = vars(parsed_args)
    create_arch_file(**kwargs)


def create_arch_file(experiment, hutch=None, path=None, dry_run=False):
    """
    Create an epicsArch file for the experiment.

    Parameter
    ---------
    experiment : str
        Experiment name, e.g.: `xpplv6818`
    hutch : str
        Hutch name if other than the one from the experiment, e.g.: `xcs`
    path : str
        Directory where to create the file if other than the default one which
        is `/cds/group/pcds/dist/pds/{}/misc/`
    dry_run : bool
        To indicate if only print to stdout the data that would be stored
        in the epicsArch file and not create the file.

    Examples
    --------
    Create the epicsArch file for experiment `xpplv6818`. If no `--path` or
    `--hutch` provided, write the file at the default path:
    `/cds/group/pcds/dist/pds/{}/misc/` whre {} is the hutch name from the
    experiment (`xpp`):

    >>> epicsarch-qs xpplv6818

    Would write: `/cds/group/pcds/dist/pds/xpp/misc/epicsArch_xpplv6818.txt`

    To change the path where you want to write the epicsArch file:

    >>> epicsarch-qs xpplv6818 --path /my/new/path/

    Would write: `/my/new/path/epicsArch_xpplv6818.txt`

    To change only the hutch name in the path:

    >>> epicsarch-qs xpplv6818 --hutch xcs

    Would write: `/cds/group/pcds/dist/pds/xcs/misc/epicsArch_xpplv6818.txt`

    To print to standard output what would be written in the epicsArch file:

    >>> epicsarch-qs xpplv6818 --dry-run
    """
    file_path = None
    if experiment and not dry_run:
        # set the path to write the epicsArch file to.
        if path:
            if path and not os.path.exists(path):
                raise IOError('Invalid path: %s' % path)
            file_path = path
        elif hutch:
            file_path = EPICS_ARCH_FILE_PATH.format(hutch.lower())
        else:
            file_path = EPICS_ARCH_FILE_PATH.format(experiment[0:3])
        create_file(exp_name=experiment, path=file_path)
    elif dry_run:
        print_dry_run(experiment)


def print_dry_run(exp_name):
    """
    Print to stdout the data that would be stored in the epicsArch file.

    Parameters
    ----------
    exp_name : str
        Experiment name, e.g.: xpplv6818

    Examples
    --------
    >>> epicsarch-qs xpplv6818 --dry-run
    """
    data = get_questionnaire_data(exp_name)
    for item in data:
        print(item)


def get_questionnaire_data(exp_name):
    """
    Return a list of items that would be written in the epicsArch file.

    Parameters
    ----------
    exp_name : str
        Experiment name, e.g.: xpplv6818

    Returns
    -------
    data_list : list
        List containing the names ans prefixes of items in the Questionnaire.
    """
    data_list = []
    items = get_items(exp_name)

    if items:
        for item in items:
            name = ''.join(('* ', item.name))
            data_list.append(name)
            data_list.append(item.prefix)
    return data_list


def get_items(exp_name):
    """
    Get all_items from Questionnaire.

    Parameters
    ----------
    exp_name : str
        Experiment name, e.g.: xpplv6818

    Returns
    -------
    items : list
        List of all_items from Questionnaire.
    """
    try:
        qs_client = get_qs_client(exp_name)
    except Exception as ex:
        logger.error('Failed to load the Questionnaire, %s', ex)
        raise ex
        sys.exit()
    items = qs_client.all_items
    if not items:
        logger.warning("No devices found in PCDS Questionnaire for %s",
                       exp_name)
        return
    return items


def create_file(exp_name, path):
    """
    Create a file with aliases and pvs from the questionnaire.

    Parameters
    ----------
    exp_name : str
        Experiment name, e.g.: `xpplv6818`
    path : str
        Directory where to create the epicsArch file.
    """
    data_list = get_questionnaire_data(exp_name)
    if not os.path.exists(path):
        raise IOError('Invalid path: %s' % path)
    exp_name = str(exp_name)
    file_path = ''.join((path, 'epicsArch_', exp_name, '.txt'))

    logger.info('Creating epicsArch file for experiment: %s', exp_name)

    with open(file_path, 'w') as f:
        for data in data_list:
            try:
                f.write(f'{data}\n')
            except OSError as ex:
                logger.error('Could not write file %s, %s', file_path, ex)


if __name__ == '__main__':
    main()



import logging

from . import utils

logger = logging.getLogger(__name__)


def get_user_objs(load, *, ask_on_failure=True):
    """
    Load the user's modules.

    All objects from these modules will be imported e.g.
    ``from module import *`` and the objects will be returned.

    Parameters
    ----------
    load : str, list of str
        The module(s) to import.

    ask_on_failure : bool, optional
        If a module fails to load, indicate what happened and ask the user if
        loading should continue.

    Returns
    -------
    objs : dict
        Mapping from object name to object
    """
    if isinstance(load, str):
        load = [load]

    objs = {}
    for module in load:
        with utils.safe_load(module):
            try:
                module_objs = utils.extract_objs(module)
                objs.update(module_objs)
            except Exception as ex:
                if not ask_on_failure:
                    raise

                utils.maybe_exit(
                    logger,
                    message=(
                        f'Devices and functions from module {module} will NOT '
                        f'be available because it failed to load with:\n\t'
                        f'{ex.__class__.__name__}: {ex}'
                    ),
                    exception_message=f'Failed to load {module}',
                )

    return objs



import json
import logging
import os
from unittest.mock import patch

import happi
import pytest
from conftest import cli_args

from ..epics_arch import create_file, get_items, main, print_dry_run

logger = logging.getLogger(__name__)


def test_epics_arch_args():
    with cli_args(['epicsarch-qs', 'xpplv6818']):
        with patch('hutch_python.epics_arch.create_file') as mock:
            main()
        mock.assert_called_once()


@patch('hutch_python.epics_arch.create_file')
def test_create_file_with_hutch_args(create_mock):
    with cli_args(['epicsarch-qs', 'xpplv6818', '--hutch', 'xpp']):
        with patch('hutch_python.epics_arch.get_items', return_value=items):
            main()
        create_mock.assert_called_once()


@patch('hutch_python.epics_arch.create_file')
@patch('os.path.exists', return_value=True)
def test_create_file_with_path_args(path_mock, create_mock):
    with cli_args(['epicsarch-qs', 'xpplv6818', '--path', '/my/path/xpp/']):
        with patch('hutch_python.epics_arch.get_items', return_value=items):
            main()
        create_mock.assert_called_once()


@patch('os.path.exists', return_value=False)
def test_create_file_with_bad_path_args(path_mock):
    with cli_args(['epicsarch-qs', 'xpplv6818', '--path', '/my/path/xpp/']):
        with pytest.raises(OSError):
            main()


@patch('hutch_python.epics_arch.get_questionnaire_data')
def test_dry_run_args(get_data_mock, items):
    with cli_args(['epicsarch-qs', 'xpplv6818', '--dry-run']):
        with patch('hutch_python.epics_arch.get_items', return_value=items):
            with patch('hutch_python.epics_arch.print_dry_run') as mock:
                main()
            mock.assert_called_once()
            get_data_mock.assert_called_once


def test_dry_run_args_exception(fake_qsbackend):
    with pytest.raises(Exception):
        get_items('somebadname')


def test_print_dry_run(items, capsys):
    with patch('hutch_python.epics_arch.get_items', return_value=items):
        expected_list = [
            '* tape_x', 'XPP:LBL:MMN:04', '* transducer_y',
            'XPP:LBL:MMN:05', '* xes_y', 'XPP:LBL:MMN:06', '* lbl',
            'XPP:USR:EVR:TRIG0', '* acromag', 'XPP:USR:ao1'
            ]
        expected_str = ''
        for m in expected_list:
            expected_str += m + '\n'

        print_dry_run('xpplv6818')
        readout = capsys.readouterr()
        assert expected_str in readout


def test_create_file(items):
    dir_path = os.path.dirname(os.path.realpath(__file__)) + '/'
    with patch('hutch_python.epics_arch.get_items', return_value=items):
        create_file('tstlr3216', path=dir_path)
        expected_file = ''.join((dir_path, 'epicsArch_tstlr3216.txt'))
        expected_list = [
            '* tape_x', 'XPP:LBL:MMN:04', '* transducer_y',
            'XPP:LBL:MMN:05', '* xes_y', 'XPP:LBL:MMN:06', '* lbl',
            'XPP:USR:EVR:TRIG0', '* acromag', 'XPP:USR:ao1']
        temp_list = [line.rstrip('\n') for line in open(expected_file)]
        # Check that the file was made
        assert os.path.exists(expected_file)
        assert temp_list == expected_list
        # Cleanup
        os.remove(expected_file)


def test_create_file_bad_path(items):
    with patch('hutch_python.epics_arch.get_items', return_value=items):
        with pytest.raises(OSError):
            create_file('tstlr3216', path='/some/bad/path/')


@pytest.fixture(scope='function')
def json_db(tmp_path_factory):
    dir_path = tmp_path_factory.mktemp("db_dir")
    json_path = dir_path / 'db.json'
    json_path.write_text(json.dumps(all_items))
    return str(json_path.absolute())


@pytest.fixture(scope='function')
def items(json_db):
    client = happi.client.Client(path=json_db)
    return client.all_items


all_items = json.loads("""{
            "tape_x": {
                "_id": "tape_x",
                "active": true,
                "args": ["{{prefix}}"],
                "beamline": "XPP",
                "kwargs": {"name": "{{name}}"},
                "lightpath": false,
                "name": "tape_x",
                "prefix": "XPP:LBL:MMN:04",
                "type": "pcdsdevices.happi.containers.LCLSItem",
                "location": "XPP goniometer",
                "purpose": "Tape X",
                "pvbase": "XPP:LBL:MMN:04",
                "stageidentity": "Newport"
                },
            "transducer_y": {
                "_id": "transducer_y",
                "active": true,
                "args": ["{{prefix}}"],
                "beamline": "XPP",
                "kwargs": {"name": "{{name}}"},
                "lightpath": false,
                "name": "transducer_y",
                "prefix": "XPP:LBL:MMN:05",
                "type": "pcdsdevices.happi.containers.LCLSItem",
                "location": "XPP goniometer",
                "purpose": "Transducer Y",
                "stageidentity": "Newport",
                "pvbase": "XPP:LBL:MMN:05"
                },
            "xes_y": {
                "_id": "xes_y",
                "active": true,
                "args": ["{{prefix}}"],
                "beamline": "XPP",
                "kwargs": {"name": "{{name}}"},
                "lightpath": false,
                "name": "xes_y",
                "prefix": "XPP:LBL:MMN:06",
                "type": "pcdsdevices.happi.containers.LCLSItem",
                "location": "XPP goniometer",
                "stageidentity": "Newport",
                "purpose": "Xes Epix Y",
                "pvbase": "XPP:LBL:MMN:06"
                },
            "lbl": {
                "_id": "lbl",
                "active": true,
                "args": ["{{prefix}}"],
                "beamline": "XPP",
                "kwargs": {"name": "{{name}}"},
                "lightpath": false,
                "name": "lbl",
                "prefix": "XPP:USR:EVR:TRIG0",
                "type": "pcdsdevices.happi.containers.LCLSItem",
                "purpose": "LBL",
                "width": "1e-5",
                "polarity": "positiv",
                "delay": "0.0008",
                "pvbase": "XPP:USR:EVR:TRIG0",
                "eventcode": "42"
                },
            "acromag": {
                "_id": "acromag",
                "active": true,
                "args": ["{{prefix}}"],
                "beamline": "XPP",
                "kwargs": {"name": "{{name}}"},
                "lightpath": false,
                "name": "acromag",
                "prefix": "XPP:USR:ao1",
                "type": "pcdsdevices.happi.containers.LCLSItem",
                "device": "Acromag IP231 16-bit",
                "pvbase": "XPP:USR:ao1",
                "purpose": "Evo Shutter 1",
                "channel": "6"
                }
                }
                """)



import logging
import os.path
from socket import gethostname
from types import SimpleNamespace

import pytest
from pcdsdaq.daq import Daq
from pcdsdaq.sim import set_sim_mode
from pcdsdaq.sim.pydaq import Control as SimControl
from pcdsdevices.interface import Presets

import hutch_python.qs_load
from hutch_python.load_conf import load, load_conf

from .conftest import (TST_CAM_CFG, BlueskyScan, ELog, QSBackend,
                       requires_elog, requires_psdaq, skip_if_win32_pcdsdaq)

logger = logging.getLogger(__name__)


@skip_if_win32_pcdsdaq
def test_file_load():
    logger.debug('test_file_load')
    set_sim_mode(True)
    objs = load(os.path.join(os.path.dirname(__file__), 'conf.yaml'))
    should_have = ('x', 'unique_device', 'calc_thing', 'daq', 'tst_beampath',
                   'scan_pvs')
    err = '{} was overriden by a namespace'
    for elem in should_have:
        assert not isinstance(objs[elem], SimpleNamespace), err.format(elem)
    assert 'tst' not in objs  # Tree namespace should be disabled
    assert len(Presets._paths) == 2


def test_exp_override():
    logger.debug('test_exp_override')
    set_sim_mode(True)
    # Should work with or without hutch name
    objs = load(os.path.join(os.path.dirname(__file__), 'conf.yaml'),
                SimpleNamespace(exp='x011'))
    assert hasattr(objs['x'], 'cats')
    objs = load(os.path.join(os.path.dirname(__file__), 'conf.yaml'),
                SimpleNamespace(exp='tstx011'))
    assert hasattr(objs['x'], 'cats')


def test_no_file():
    logger.debug('test_no_file')
    objs = load()
    assert len(objs) > 1


def test_conf_empty():
    logger.debug('test_conf_empty')
    objs = load_conf({})
    assert len(objs) > 1


@requires_elog
def test_elog(monkeypatch, temporary_config):
    logger.debug('test_elog')
    monkeypatch.setattr(hutch_python.load_conf, 'HutchELog', ELog)
    # No platform
    objs = load_conf({'hutch': 'TST'})
    assert objs['elog'].station is None
    # Check authentication worked correctly
    assert objs['elog'].user == 'user'
    assert objs['elog'].pw == 'pw'
    # Define default platform
    objs = load_conf({'daq_platform': {'default': 1},
                      'hutch': 'TST'})
    assert objs['elog'].station is None
    # Define host platform
    hostname = gethostname()
    objs = load_conf({'daq_platform': {'default': 3,
                                       hostname: 4},
                      'hutch': 'TST'})
    assert objs['elog'].station == '1'


@requires_psdaq
def test_lcls2_daq_config(dummy_zmq_lcls2):
    logger.debug('test_lcls2_daq')

    host = 'fake-hostname-drp'
    platform = 1
    config = {
        'daq_type': 'lcls2',
        'daq_host': host,
        'daq_platform': {'default': platform},
    }
    objs = load_conf(config)
    daq = objs['daq']
    assert isinstance(daq, BlueskyScan)
    assert daq.control.host == host
    assert daq.control.platform == platform


@skip_if_win32_pcdsdaq
def test_simdaq_config():
    logger.debug('test_simdaq_config')
    objs = load_conf({'daq_type': 'lcls1-sim'})
    daq = objs['daq']
    assert isinstance(daq, Daq)
    daq.connect()
    assert isinstance(daq._control, SimControl)


def test_nodaq_config():
    logger.debug('test_nodaq_config')
    objs = load_conf({'daq_type': 'nodaq'})
    with pytest.raises(KeyError):
        objs['daq']


def test_camviewer_load(monkeypatch):
    logger.debug('test_camviewer_load')
    monkeypatch.setattr(hutch_python.load_conf, 'CAMVIEWER_CFG', TST_CAM_CFG)
    objs = load_conf({'hutch': ''})
    assert 'camviewer' in objs
    assert 'my_cam' in dir(objs['camviewer'])


def test_skip_failures():
    logger.debug('test_skip_failures')
    # Should not raise
    load_conf(dict(hutch=345243, db=12351324, experiment=2341234, load=123454,
                   bananas='dole'))


def test_auto_experiment(fake_curexp_script):
    logger.debug('test_auto_experiment')
    hutch_python.qs_load.QSBackend = QSBackend
    objs = load_conf(dict(hutch='tst'))
    assert objs['inj_x'].run == '15'
    assert objs['inj_x'].proposal == 'LR12'
    assert objs['x'].inj_x == objs['inj_x']


def test_cannot_auto():
    logger.debug('test_cannot_auto')
    # Fail silently
    load_conf(dict(hutch='tst'))



"""
This module is used to both gather and report information for the purpose of
identifying bugs
"""
import os
import getpass
import logging
import textwrap
import tempfile
import warnings
import subprocess
from configparser import ConfigParser, NoOptionError, NoSectionError

import requests
import simplejson
from IPython.core.history import HistoryAccessor
from IPython.core.magic import Magics, magics_class, line_magic
from IPython.utils.io import capture_output
from jinja2 import Environment, PackageLoader

from .log_setup import get_session_logfiles

logger = logging.getLogger(__name__)


# This function is aliased here so it can be easily be patched by the test
# suite. This allows us to eloquently simulate user inputted data
request_input = input


def get_current_environment():
    """
    Get the current CONDA environment

    The environment name is gathered by first checking environment variable
    ``$CONDA_ENVNAME`` that is set by the hutch-python startup scripts. If for
    whatever reason that is not available we check ``$CONDA_DEFAULT_ENV`` which
    is set by Conda itself. The reason this is not relied on primarily is it
    has a strange name and is entirely undocumented but seems to work
    effectively.

    In addition the hutch-python startup script sets the ``PYTHONPATH`` to pick
    up packages in "development" mode. The list of package names installed this
    way is found to help inform how the current Python environment differs from
    the enforced CONDA environment

    Returns
    -------
    env: str
        Name of environment

    dev_pkgs :list
        List of packages installed in the development folder
    """
    # Search for the environment variable set by the hutch python setup
    env = os.getenv('CONDA_ENVNAME')
    # Otherwise look for built-in Conda environment variables
    if not env:
        env = os.getenv('CONDA_DEFAULT_ENV')
    # Check the top level PYTHONPATH to see if we have packages installed in
    # development mode
    dev = os.getenv('PYTHONPATH')
    if dev:
        try:
            dev_pkgs = os.listdir(dev)
        except FileNotFoundError:
            logger.debug("No dev folder found")
            dev_pkgs = list()
    else:
        dev_pkgs = list()
    return env, dev_pkgs


def get_last_n_commands(n):
    """
    Find the last n commands entered in the IPython session

    Parameters
    ----------
    n : int
        Number of commands to retrieve

    Returns
    -------
    commands : str
        Block of text representing user input
    """
    # Ignore warnings generated by the HistoryAccessor. This can be removed
    # when https://github.com/ipython/ipython/pull/11054 reaches our release
    # environment
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', UserWarning)
        ha = HistoryAccessor()
    # Parse the last n commands of the IPython history, joining printed
    # messages
    return '\n'.join([cmd[2] for cmd in ha.get_tail(n, include_latest=True)])


def get_text_from_editor():
    """
    Request a description written in a text editor

    Opens a vim session with a prompt to request a detailed response from the
    operator.

    Returns
    -------
    text : str
        Block of text of the user input
    """
    with tempfile.NamedTemporaryFile(suffix='.tmp', mode='w+t') as f:
        # Create a temporary file with instructions on describing bug
        f.write(message + '\n\n')
        f.flush()
        # Open the editor and allow the user to type
        editor = os.environ.get('EDITOR', 'vim')
        subprocess.call([editor, f.name])
        # Read and clean the file
        f.seek(0)
        text = ''.join([line.lstrip() for line in f.readlines()
                        if line and not line.lstrip().startswith('#')])
        return '\n'.join(textwrap.wrap(text, width=100))


def report_bug(title=None, description=None, author=None,
               prior_commands=None, captured_output=None, **kwargs):
    """
    Report a bug from the IPython session

    The purpose of this command is to collect the necessary information to
    later help diagnose and troubleshoot the issue. This is written as an
    interactive tool, but it can be used in a non-interactive way by entering
    the information on the call.

    By the end we should have gathered:

        * A brief description of the problem
        * The relevant commands to the bug report
        * Name of the current CONDA environment
        * Any packages installed in "development" mode
        * Relevant logfiles
        * Name of bug report author

    Parameters
    ----------
    title : str, optional
        One sentence description of the issue

    description : str, optional
        Written description of problem. If this is not provided, a text editor
        is launched that request the information from the user

    author : str, optional
        Name of bug report author. If not provided, this is requested from the
        user via command line

    prior_commands : int, optional
        Number of prior commands to capture. If this is not provided, this is
        requested from the user via command line.

    captured_output : str, optional
        Captured output from the command

    kwargs:
        Pass authentication information to :func:`.post_to_github`
    """
    logger.debug("Reporting a bug from the IPython terminal ...")
    if not title:
        title = request_input('Please provide a one sentence description of '
                              'the problem you are encountering: ')
    # Grab relevant commands
    if not prior_commands:
        try:
            n = request_input("How many of the previous commands are relevant "
                              "to the issue you would like investigated?: ")
            prior_commands = int(n)
        except ValueError:
            logger.error("Invalid input %s", n)
            # Only select the last command by default
            prior_commands = 1
    # Grab specified number of commands
    try:
        commands = get_last_n_commands(prior_commands)
    # If the user somehow has no IPython history this is raised. A very rare
    # occurence except for in Continuous Integration tests
    except OSError:
        logger.exception('Unable to retrieve commmands from the '
                         'IPython session')
        commands = ''
    # Get a more specific description
    description = description or get_text_from_editor()
    # Find the author
    if not author:
        author = request_input('Please enter a name so we can follow-up with '
                               'additional questions: ')
        author = author or getpass.getuser()
    # Gather environment information
    conda_env, dev_pkgs = get_current_environment()
    # Gather logfiles
    logfiles = get_session_logfiles()
    # Save the report to JSON
    return post_to_github({'title': title, 'author': author,
                           'commands': commands, 'description': description,
                           'env': conda_env, 'logfiles': logfiles,
                           'output': captured_output, 'dev_pkgs': dev_pkgs},
                          **kwargs)


def post_to_github(report, user=None, pw=None, proxies=None):
    """
    Post an issue report to GitHub

    Authentication can be done in three different ways depending on preference.
    First, the call can be made with the username and password specified. If
    this is not done we first look for a configuration file web.cfg that has
    a section labeled GitHub which looks like:

    .. code:: ini

        [GITHUB]
        user=username
        pw=password
        proxy=http://proxyhost:port

    If this is not available the username and password will be requested via
    the command line. The proxy specification allows posts from hosts without
    direct connection to the internet. Please consult PCDS for information
    about available hosts and ports.

    Parameters
    ----------
    report: dict
        A report dictionary with keys:

            * title
            * author
            * commands
            * description
            * env
            * logfiles
            * output
            * dev_pkgs

    user: str, optional
        Username of GitHub profile.

    pw : str, optional
        Password for GitHub profile. This will be queried for if not provided
        in the function call.

    proxies : dict, optional
        Mapping of protocol to hostname and port.
    """
    proxies = proxies or dict()
    # Determine authentication method. No username or password search for
    # configuration file with GITHUB section
    if not user and not pw:
        # Find configuration file
        cfg = ConfigParser()
        cfgs = cfg.read(['web.cfg', '.web.cfg',
                         os.path.expanduser('~/.web.cfg'),
                         'qs.cfg', '.qs.cfg',
                         os.path.expanduser('~/.qs.cfg')])
        if cfgs:
            # Grab login information
            try:
                user = cfg.get('GITHUB', 'user')
                pw = cfg.get('GITHUB', 'pw')
            except (NoOptionError, NoSectionError):
                logger.debug('No GITHUB section in configuration file '
                             'with user and pw entries')
            # Grab proxy information if we will be using web.cfg
            if (user or pw) and not proxies:
                try:
                    proxy_name = cfg.get('GITHUB', 'proxy')
                    logger.debug("Using proxy host %s", proxy_name)
                    proxies = {'https': proxy_name}
                except NoOptionError:
                    logger.debug("No proxy information found")
        # No valid configurations
        else:
            logger.debug('No "web.cfg" file found')
    # Manually ask if we didn't get the username or password already
    if not user:
        user = input('Github Username: ')
    if not pw:
        pw = getpass.getpass('Password for GitHub Account {}: '
                             ''.format(user))
    # Our url to create issues via POST
    url = 'https://api.github.com/repos/pcdshub/Bug-Reports/issues'
    # Create the body of the template
    env = Environment(loader=PackageLoader('hutch_python'),
                      trim_blocks=True, lstrip_blocks=True)
    template = env.get_template('issue.template')
    body = template.render(report)
    # Requests session
    session = requests.Session()
    session.auth = (user, pw)
    session.proxies.update(proxies)
    issue = {'title': report['title'],
             'body': body,
             'assignee': None,
             'milestone': None,
             'labels': []}  # TODO: Determine hutch to create issue for
    # Post to GitHub
    r = session.post(url, simplejson.dumps(issue))
    if r.status_code == 201:
        logger.info("Succesfully created GitHub issue")
    else:
        logger.exception("Could not create GitHub issue. HTTP Status Code: %s",
                         r.status_code)


@magics_class
class BugMagics(Magics):
    """Magics function for report_bug function"""

    @line_magic
    def report_bug(self, line):
        """Creates a bug_report while running the given line"""
        # Store the output of our command
        with capture_output() as shell_output:
            self.shell.run_cell(line)
        # Show the capture output to the user
        shell_output.show()
        # Create the report
        report_bug(prior_commands=1, captured_output=shell_output.stdout)


message = """\
# Please describe the issue you are wishing to report. What did you expect to
# happen? What actually happened? Does this issue occur every time you use this
# function?

# Lines that start with a '#' will be ignored. Save the session and exit to
# store the description. Press "i" to be begin typing, then "Esc" followed by
# ":wq" to exit.
"""


def load_ipython_extension(ipython):
    ipython.register_magics(BugMagics)



"""
This module is responsible for accumulating all loaded objects and making sure
they are available in the ``xxx.db`` virtual module. It is used extensively in
`load_conf.load_conf`.
"""
import datetime
import logging
import sys
import textwrap
from importlib import import_module
from pathlib import Path

from .utils import HelpfulNamespace

logger = logging.getLogger(__name__)


class LoadCache:
    """
    Class that accumulates objects in a virtual module.

    This virtual module can be imported from as if it were a normal module.

    Parameters
    ----------
    module: ``str``
        Name of the virtual module to create. If the module has the same
        name as a real module, the real module will be masked.

    hutch_dir: ``Path``, optional
        This allows us to write a ``db.txt`` file to let the user know what
        objects get imported.

    **objs: kwargs
        Initial objects to place into the namespace

    Attributes
    ----------
    objs: `HelpfulNamespace`
        This is a namespace containing all the objects that have been attached
        to the ``LoadCache``.
    """
    def __init__(self, module, hutch_dir=None, **objs):
        self.objs = HelpfulNamespace(**objs)
        self.hutch_dir = hutch_dir
        self.module = module
        self.spoof_module(module)
        self.spoof_module('hutch_python.db')

    def spoof_module(self, module_name):
        """
        Create a fake module that is actually self.objs
        """
        # Check for real module that it needs to be slipped into
        module_parts = module_name.split('.')
        parent = '.'.join(module_parts[:-1])
        if parent:
            try:
                parent_module = import_module(parent)
                setattr(parent_module, module_parts[-1], self.objs)
            except ImportError:
                logger.debug('Skip patching parent module %s, does not import',
                             parent, exc_info=True)

        # Place it here so it looks like we've already imported it
        sys.modules[module_name] = self.objs

    def __call__(self, **objs):
        """
        Add objects to the namespace.

        Parameters
        ----------
        **objs: kwargs
            The key will is the namespace-accessible name, and the object
            is the object we are adding.
        """
        self.objs.__dict__.update(**objs)

    def write_file(self):
        """
        Write a ``db.txt`` file in the hutch's directory. This file informs
        the user which objects get loaded by ``hutch-python``.
        """
        if self.hutch_dir is not None:
            parts = self.module.split('.')
            parts[-1] = parts[-1] + '.txt'
            db_path = self.hutch_dir / Path('/'.join(parts))
            text = (header.format(parts[0])
                    + body.format(datetime.datetime.now()))
            for name, obj in self.objs.__dict__.items():
                text += '{:<20} {}\n'.format(name, obj.__class__)
            if not db_path.exists():
                db_path.touch()
                db_path.chmod(0o666)
            with db_path.open('w') as f:
                f.write(text)

    def doc(self, **docs):
        """
        Add docstrings to the cached objects.

        Parameters
        ----------
        **objs: kwargs
            The key should be the object name, the value should be
            documentation about that object and why it is included.
        """
        for key, value in docs.items():
            obj = self.objs[key]
            if obj.__doc__:
                obj.__doc__ = value + '\n' + textwrap.dedent(obj.__doc__)
            else:
                obj.__doc__ = value


# For writing the files
header = """
The objects referenced in this file are populated by the {0}python\n
initialization. If you wish to use devices from this file, import\n
them from {0}.db after calling the {0}python startup script.\n\n
"""
body = ('hutch-python last loaded on {}\n'
        'with the following objects:\n\n')



"""
Module to contain ophyd-specific settings that must be applied prior to
creating any ophyd objects.
"""
import os

from ophyd.signal import EpicsSignalBase

# Environment variable definitions
CONN_VAR = 'HUTCH_PYTHON_CONNECTION_TIMEOUT'
READ_VAR = 'HUTCH_PYTHON_READ_TIMEOUT'
WRITE_VAR = 'HUTCH_PYTHON_WRITE_TIMEOUT'
AUTO_VAR = 'HUTCH_PYTHON_AUTO_MONITOR'

# Acceptable values to make auto_monitor True
AUTO_TRUE = {'true', 't', 'yes', 'y', '1'}


def setup_ophyd():
    """
    Call EpicsSignalBase.set_defaults with pcds defaults.

    First, checks the following environment variables to set:
      - HUTCH_PYTHON_CONNECTION_TIMEOUT -> connection_timeout
      - HUTCH_PYTHON_READ_TIMEOUT -> timeout
      - HUTCH_PYTHON_WRITE_TIMEOUT -> write_timeout
      - HUTCH_PYTHON_AUTO_MONITOR -> auto_monitor

    If missing, default to the following behavior:

    Sets the following to the current ophyd defaults:
      - connection_timeout
      - timeout (read timeout)
      - auto_monitor

    Overriding defaults with themselves seems redundant, but I want to
    be insulated from changes to the upstream consensus.

    Sets a new default to the following:
      - write timeout (we don't want to wait forever)
    """
    EpicsSignalBase.set_defaults(
        connection_timeout=float(os.environ.get(CONN_VAR, 1.0)),
        timeout=float(os.environ.get(READ_VAR, 2.0)),
        write_timeout=float(os.environ.get(WRITE_VAR, 5.0)),
        auto_monitor=os.environ.get(AUTO_VAR, 'false').lower() in AUTO_TRUE,
        )



import logging

import IPython.lib.pretty as pretty
import pytest

from hutch_python import utils

logger = logging.getLogger(__name__)
_TEST = 4


def test_safe_load():
    logger.debug('test_safe_load')

    with utils.safe_load('zerodiv'):
        1/0

    with utils.safe_load('apples', cls='fruit'):
        apples = 4

    assert apples == 4


def test_get_current_experiment(fake_curexp_script):
    logger.debug('test_get_current_experiment')
    assert utils.get_current_experiment('tst') == 'tstlr1215'


def test_iterable_namespace():
    logger.debug('test_iterable_namespace')

    ns = utils.IterableNamespace(a=1, b=2, c=3)

    assert list(ns) == [1, 2, 3]
    assert len(ns) == 3


def test_helpful_namespace_pretty_print():
    class Obj:
        """Docstring"""

    ns = utils.HelpfulNamespace(obj_a=Obj(), obj_b=Obj())
    pretty_ns = pretty.pretty(ns)
    print("Pretty repr is", pretty_ns)
    assert "obj_a" in pretty_ns
    assert "obj_b" in pretty_ns
    assert "Docstring" in pretty_ns


def test_helpful_namespace_html_print():
    class Obj:
        """Docstring"""

    ns = utils.HelpfulNamespace(obj_a=Obj(), obj_b=Obj())
    html_ns = ns._repr_html_()
    assert "<td>obj_a</td>" in html_ns
    assert "<td>obj_b</td>" in html_ns
    assert "<td>Docstring</td>" in html_ns


def test_count_leaves():
    logger.debug('test_count_leaves')

    ns0 = utils.IterableNamespace(a=utils.IterableNamespace())
    ns1 = utils.IterableNamespace(a=1, b=utils.IterableNamespace())
    ns2 = utils.IterableNamespace(a=utils.IterableNamespace(a=1),
                                  b=utils.IterableNamespace(b=2))
    ns3 = utils.IterableNamespace(a=1,
                                  b=utils.IterableNamespace(a=1, b=2))

    assert utils.count_ns_leaves(ns0) == 0
    assert utils.count_ns_leaves(ns1) == 1
    assert utils.count_ns_leaves(ns2) == 2
    assert utils.count_ns_leaves(ns3) == 3


def test_extract_objs():
    logger.debug('test_extract_objs')
    # Has no __all__ keyword
    objs = utils.extract_objs('sample_module_1')
    assert objs['hey'] == '4horses'
    assert objs['milk'] == 'cows'
    assert objs['some_int'] == 5
    # Has an __all__ keyword
    objs = utils.extract_objs('sample_module_2.py')
    assert objs == dict(just_this=5.0)
    # Takes a list
    objs = utils.extract_objs(['sample_module_1', 'sample_module_2'])
    assert len(objs) == 5
    # Called with no scope, no skip hidden
    objs = utils.extract_objs(skip_hidden=False)
    assert objs['_TEST'] == 4


def test_find_class():
    logger.debug('test_find_class')
    # Find some standard type that needs an import
    found_Request = utils.find_class('urllib.request.Request')
    from urllib.request import Request
    assert found_Request is Request
    # Find some built-in type
    found_float = utils.find_class('float')
    assert found_float is float
    # Raises error if nothing is found
    with pytest.raises(ImportError):
        utils.find_class('aseoiajsdf')


def test_strip_prefix():
    logger.debug('test_strip_prefix')
    assert utils.strip_prefix('cats_dogs', 'cats') == 'dogs'
    assert utils.strip_prefix('cats', 'dogs') == 'cats'


def test_hutch_banner():
    logger.debug('test_hutch_banner')
    utils.hutch_banner()
    utils.hutch_banner('mfx')



"""
Module that contains general-use utilities. Some of these are useful outside of
``hutch-python``, while others are used in multiple places throughout the
module.
"""
import functools
import inspect
import logging
import os
import socket
import sys
import time
from contextlib import contextmanager
from functools import partial
from importlib import import_module
from subprocess import check_output
from types import SimpleNamespace

import prettytable
import pyfiglet

from .constants import (CLASS_SEARCH_PATH, CUR_EXP_SCRIPT, HUTCH_COLORS,
                        SUCCESS_LEVEL)

logging.addLevelName('SUCCESS', SUCCESS_LEVEL)
logger = logging.getLogger(__name__)
logger.success = partial(logger.log, SUCCESS_LEVEL)


@contextmanager
def safe_load(name, cls=None):
    """
    Context manager to safely run a block of code.

    This will abort running code and resume the rest of the program if
    something fails. This can be used to wrap user code with unknown behavior.
    This will log standard messages to indicate success or failure.

    Parameters
    ----------
    name: ``str``
        The name of the load to be logged. This will be used in the log
        message.

    cls: ``type``, optional
        The class of a loaded object to be logged. This will be used in the log
        message.
    """
    start_time = time.monotonic()

    if cls is None:
        identifier = name
    else:
        identifier = ' '.join((name, str(cls)))
    logger.info('Loading %s...', identifier)
    try:
        yield
        duration = time.monotonic() - start_time
        logger.success('Successfully loaded %s in %.2f s',
                       identifier, duration)
    except Exception as exc:
        duration = time.monotonic() - start_time
        logger.error('Failed to load %s after %.2f s', identifier, duration)
        logger.debug(exc, exc_info=True)


def get_current_experiment(hutch):
    """
    Get the current experiment for ``hutch``.

    This currently works by running an external script on NFS, but this will be
    changed in the future.

    Parameters
    ----------
    hutch: ``str``
        The hutch we would like to know the current experiment of

    Returns
    -------
    expname: ``str``
        Full experiment name, e.g. ``xppls2516``
    """
    script = CUR_EXP_SCRIPT.format(hutch)
    return check_output(script.split(' '), universal_newlines=True).strip('\n')


class HelpfulNamespace(SimpleNamespace):
    """
    ``SimpleNamespace`` that can be iterated over, with a fancy table repr.

    This means we can call funtions like ``list`` on these objects to see all
    of their contents, we can put them into ``for loops``, and we can use them
    in ``generator expressions``.

    This class also has the added feature where ``len`` will correctly tell you
    the number of objects in the ``namespace``.
    """
    _ignore_attrs = {"__doc__"}
    _ignore_underscore = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__doc__ = self._get_docstring()

    def _get_items(self):
        """
        Get all items contained in this namespace, sorted by attribute name.

        Yields
        ------
        attr : str
            The attribute name.
        obj : object
            The object associated with attr. i.e., ``self.{attr}``
        """
        # Sorts alphabetically by key
        for attr, obj in sorted(self.__dict__.items()):
            if attr not in self._ignore_attrs:
                if not self._ignore_underscore or not attr.startswith("_"):
                    yield attr, obj

    def __iter__(self):
        for _, obj in self._get_items():
            yield obj

    def __len__(self):
        return len(list(self._get_items()))

    def __getitem__(self, item):
        return self.__dict__[item]

    def _get_docstring(self):
        table = self._as_table_()
        if table.rowcount == 0:
            return ""
        return str(table)

    def _as_table_(self, *, nest_html=False):
        """
        Represent the namespace as a PrettyTable.

        Parameters
        ----------
        nest_html : bool, optional
            For nested tables, use HTML if set, or plain ASCII text if not.
        """
        table = prettytable.PrettyTable()
        table.add_column("Attribute", [])
        table.add_column("Class", [])
        table.add_column("Description", [], align="l")
        multiline_rows = False
        for attr, obj in self._get_items():
            if attr.startswith('_'):
                continue
            docs = inspect.getdoc(obj) or ""
            if isinstance(obj, HelpfulNamespace):
                # TODO: in the future, we can try to embed a table. However,
                # prettytable will escape the HTML and cause it to render
                # as &lt;tr&gt; instead of <tr>. Oh well.
                # docs = obj._as_table_(nest_html=True).get_html_string()
                multiline_rows = True
                # Full docstring from the sub-namespace will be included.
            elif docs:
                # For everything else, include just the first line
                docs = docs.splitlines()[0]
            table.add_row([attr, type(obj).__name__, docs])
        if multiline_rows:
            table.hrules = prettytable.ALL
        return table

    def _repr_html_(self):
        """This is an IPython hook for returning the html representation."""
        table = self._as_table_(nest_html=True)
        if table.rowcount == 0:
            return (
                f"This {type(self).__name__} has no available attributes.<br/>"
            )
        return f"""
This {type(self).__name__} has the following attributes available:
<br/>
{table.get_html_string()}
"""

    def _repr_pretty_(self, pretty, cycle):
        """This is an IPython hook for returning an ASCII representation."""
        table = self._as_table_()
        if table.rowcount == 0:
            pretty.text(f"""\
This {type(self).__name__} has no available attributes.
""")
        else:
            try:
                table.max_table_width = os.get_terminal_size()[0]
            except OSError:
                # This means we aren't actually in a terminal
                pass
            pretty.text(f"""\
This {type(self).__name__} has the following attributes available:

{table}
""")


# Back-compat; it's extra helpful now!
IterableNamespace = HelpfulNamespace


def count_ns_leaves(namespace):
    """
    Count the number of objects in a nested `IterableNamespace`.

    Given an `IterableNamespace` that contains other `IterableNamespace`
    objects that may in themselves contain `IterableNamespace` objects,
    determine how many non-`IterableNamespace` objects are in the tree.
    """
    count = 0
    for obj in namespace:
        if isinstance(obj, IterableNamespace):
            count += count_ns_leaves(obj)
        else:
            count += 1
    return count


def extract_objs(scope=None, skip_hidden=True, stack_offset=0):
    """
    Return all objects with the ``scope``.

    This can be though of as a ``*`` import, and it obeys the ``__all__``
    keyword functionality.

    Parameters
    ----------
    scope: ``module``, ``namespace``, or ``list`` of these, optional
        If provided, we'll import from this object.
        If omitted, we'll include all objects that have been loaded by
        hutch_python and everything in the caller's global frame.

    skip_hidden: ``bool``, optional
        If ``True``, we'll omit objects with leading underscores.

    stack_offset: ``int``, optional
        If ``scope`` was not provided, we'll use ``stack_offset`` to determine
        which frame is the user's frame. Leave this at zero if you want the
        objects in the caller's frame, and increase it by one for each level
        up the stack your frame is.

    Returns
    -------
    objs: ``dict``
        Mapping from name in scope to object
    """
    if scope is None:
        stack_depth = 1 + stack_offset
        frame = sys._getframe(stack_depth)
        try:
            objs = extract_objs(scope='hutch_python.db',
                                skip_hidden=skip_hidden,
                                stack_offset=stack_offset)
        except ImportError:
            objs = {}
        objs.update(frame.f_globals)
    else:
        if isinstance(scope, list):
            objs = {}
            for s in scope:
                objs.update(extract_objs(scope=s,
                                         skip_hidden=skip_hidden,
                                         stack_offset=stack_offset))
        else:
            if isinstance(scope, str):
                if scope.endswith('.py'):
                    scope = scope[:-3]
                scope = import_module(scope)
            objs = scope.__dict__.copy()

    all_kwd = objs.get('__all__')
    if all_kwd is None:
        if skip_hidden:
            return {k: v for k, v in objs.items() if k[0] != '_'}
        else:
            return objs
    else:
        all_objs = {}
        for kwd in all_kwd:
            all_objs[kwd] = objs.get(kwd)
        return all_objs


def find_object(obj_path):
    """
    Given a string module path to an object, return that object.

    Parameters
    ----------
    obj_path: ``str``
        String module path to an object

    Returns
    -------
    obj: ``object``
        That object
    """
    parts = obj_path.split('.')
    module_path = '.'.join(parts[:-1])
    class_name = parts[-1]
    module = import_module(module_path)
    return getattr(module, class_name)


def find_class(class_path, check_defaults=True):
    """
    Find a ``type`` object given a ``str``.

    Given a string class name, either return the matching built-in type or
    import the correct module and return the type.

    Parameters
    ----------
    class_path: ``str``
        Built-in type name or import path e.g. ``ophyd.device.Device``

    check_defaults: ``bool``
        If ``True``, try checking inside each module in ``CLASS_SEARCH_PATH``

    Returns
    -------
    cls: ``type``
        The class we found
    """
    try:
        if '.' in class_path:
            return find_object(class_path)
        else:
            return eval(class_path)
    except NameError:
        if check_defaults:
            for default in CLASS_SEARCH_PATH:
                try:
                    return find_class(default + '.' + class_path,
                                      check_defaults=False)
                except AttributeError:
                    pass
        raise ImportError('Could not find_class for {}'.format(class_path))


def strip_prefix(name, strip_text):
    """
    Strip the first section of an underscore-separated ``name``.

    If the first section matches the ``strip_text``, we'll remove it.
    Otherwise, the object will remain unchanged.

    Parameters
    ----------
    name: ``str``
        underscore_separated_name to strip from

    strip_text: ``str``
        Text to strip from the name, if it matches the first segment

    Returns
    -------
    stripped: ``str``
        The ``name``, modified or unmodified.
    """
    if name.startswith(strip_text):
        return name[len(strip_text)+1:]
    else:
        return name


def maybe_exit(logger, message, exception_message, *, exit_code=1):
    """
    For potentially fatal exceptions, prompt the user whether or not to exit.

    This outputs a full exception traceback first with the message
    `exception_message`, before a friendlier `message`, and then querying the
    user.

    Parameters
    ----------
    logger : logging.Logger
        The logger instance to output a message.

    message : str
        The user-friendly error message.

    exception_message : str
        The error to show along with the full exception traceback.

    exit_code : int, optional
        The exit code, if the user decides to exit.
    """
    logger.exception(exception_message)
    logger.error(message)

    response = input('Continue loading hutch-python? [Yn] ')
    if response.lower() not in {'y', ''}:
        sys.exit(exit_code)


def hutch_banner(hutch_name='Hutch '):
    """
    Display the hutch's banner.

    Parameters
    ----------
    hutch_name: ``str``
        Name of the hutch to produce a banner for.
    """
    text = hutch_name + 'Python'
    f = pyfiglet.Figlet(font='big')
    banner = f.renderText(text)
    if hutch_name in HUTCH_COLORS:
        banner = '\x1b[{}m'.format(HUTCH_COLORS[hutch_name]) + banner
    print(banner)


@functools.lru_cache(maxsize=1)
def get_fully_qualified_domain_name():
    """Get the fully qualified domain name of this host."""
    try:
        return socket.getfqdn()
    except Exception:
        logger.warning(
            "Unable to get machine name.  Things like centralized "
            "logging may not work."
        )
        logger.debug("getfqdn failed", exc_info=True)
        return ""



from setuptools import find_packages, setup

import versioneer

with open("requirements.txt", "rt") as fp:
    install_requires = [
        line for line in fp.read().splitlines()
        if line and not line.startswith("#")
    ]

setup(
    name="hutch-python",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license="BSD",
    author="SLAC National Accelerator Laboratory",
    packages=find_packages(),
    include_package_data=True,
    description=(
        "Launcher and Config Reader for LCLS Interactive IPython Sessions"
    ),
    entry_points={
        "console_scripts": [
            "hutch-python=hutch_python.cli:main",
            "epicsarch-qs=hutch_python.epics_arch:main",
        ]
    },
    install_requires=install_requires,
    python_requires=">=3.6",
)



"""
This module is responsible for reading and interpreting the ``conf.yml`` file.
The file's specification can be found on the `yaml_files` page.
"""
import logging
from copy import copy
from pathlib import Path
from socket import gethostname

from . import mpl_config  # noqa: F401

import matplotlib
import yaml
from archapp.interactive import EpicsArchive
from bluesky import RunEngine
from bluesky.callbacks.best_effort import BestEffortCallback
from bluesky.callbacks.core import LiveTable
from bluesky.callbacks.mpl_plotting import initialize_qt_teleporter
from pcdsdaq.daq import Daq
from pcdsdaq.scan_vars import ScanVars
from pcdsdaq.sim import set_sim_mode as set_daq_sim
from pcdsdevices.interface import setup_preset_paths


from . import calc_defaults, plan_defaults, sim, log_setup
from .cache import LoadCache
from .cam_load import read_camviewer_cfg
from .constants import CAMVIEWER_CFG, VALID_KEYS
from .debug import load_debug
from .exp_load import get_exp_objs
from .happi import get_happi_objs, get_lightpath
from .lcls import global_devices, global_device_docs
from .namespace import class_namespace
from .ophyd_settings import setup_ophyd
from .options import load_options
from .qs_load import get_qs_objs
from .user_load import get_user_objs
from .utils import (get_current_experiment, hutch_banner, safe_load,
                    HelpfulNamespace)

try:
    from elog import HutchELog
except ImportError:
    HutchELog = None

logger = logging.getLogger(__name__)


def load(cfg=None, args=None):
    """
    Read the config file and convert the yaml format into a ``dict``.

    This method:

    - Finds the hutch's launch directory
    - Modify the conf if specified by args
      - ``exp`` is an override for the experiment key
    - Loads the hutch's objects by calling `load_conf.load_conf`

    Parameters
    ----------
    cfg: ``str``, optional
        Path to the ``conf.yml`` file.
        If this is missing, we'll end up with a very empty environment.

    args: ``Namespace``, optional
        All of the arguments from the cli.

    Returns
    -------
    objs: ``dict{str: object}``
        All objects defined by ``conf.yml``. The strings are the names
        that will be accessible in the global namespace.
    """
    if cfg is None:
        conf = {}
        hutch_dir = None
    else:
        with open(cfg, 'r') as f:
            conf = yaml.safe_load(f)
        conf_path = Path(cfg)
        hutch_dir = conf_path.parent

    if args is not None and args.exp is not None:
        logger.debug('forcing experiment=%s', args.exp)
        conf['experiment'] = args.exp

    return load_conf(conf, hutch_dir=hutch_dir, args=args)


def load_conf(conf, hutch_dir=None, args=None):
    """
    Step through the object loading procedure, given a configuration.

    The procedure is:

    - Check the configuration for errors
    - Display the banner by calling `hutch_banner`
    - Use ``hutch`` key to create ``hutch.db`` importable namespace to
      stash the objects. This will be literally ``hutch.db`` if hutch is
      not provided, or the hutch name e.g. ``mfx.db``.
    - Load debug tools
    - Load options
    - Set ophyd signal default timeouts
    - Create a ``RunEngine``, ``RE``
    - Import ``plan_defaults`` and include as ``p``, ``plans``
    - Create a ``daq`` object with ``RE`` registered.
    - Create a ``scan_pvs`` object, and ``enable`` it.
    - Use ``hutch`` and ``daq_platform`` keys to create the ``elog`` object
      and configure it to match the correct experiment.
    - Use ``db`` key to load devices from the ``happi`` beamline database
      and create a ``hutch_beampath`` object from ``lightpath``
    - Use ``hutch`` key to load detector objects from the ``camviewer``
      configuration file.
    - Use ``experiment`` key to select the current experiment

        - If ``experiment`` was missing, autoselect experiment using
          ``hutch`` key

    - Use current experiment to load experiment objects from questionnaire
    - Use ``load`` key to bring up the user's ``beamline`` modules
    - Use current experiment to load experiment file

    If a conf key is missing, we'll note it in a ``logger.info`` message.
    If an extra conf entry is found, we'll note it in a ``logger.warning``
    message.
    If an automatically selected file is missing, we'll note it in a
    ``logger.info`` message.
    All other errors will be noted in a logger.error message.

    Parameters
    ----------
    conf : dict
        ``dict`` interpretation of the original yaml file

    hutch_dir : ``Path`` or ``str``, optional
        ``Path`` object that points to the hutch's launch directory. This is
        the directory that includes the ``experiments`` directory and a
        hutchname directory e.g. ``mfx``
        If this is missing, we'll be unable to write the ``db.txt`` file,
        do relative filepath database selection for ``happi``,
        or establish a preset positions directory.

    args : argparse.Namespace, optional
        The namespace returned from the cli argument parsing, or None

    Returns
    ------
    objs: ``dict{str: object}``
        See the return value of `load`
    """
    # Warn user about excess config entries
    for key in conf:
        if key not in VALID_KEYS:
            txt = ('Found %s in configuration, but this is not a valid key. '
                   'The valid keys are %s')
            logger.warning(txt, key, VALID_KEYS)

    # Grab configurations from dict, set defaults, show missing
    try:
        hutch = conf['hutch']
        if isinstance(hutch, str):
            hutch = hutch.lower()
        else:
            logger.error('Invalid hutch conf %s, must be string.', hutch)
            hutch = None
    except KeyError:
        hutch = None
        logger.info(('Missing hutch from conf. Will skip elog '
                     'and cameras.'))

    # Display the banner
    if hutch is None:
        hutch_banner()
    else:
        hutch_banner(hutch)

    try:
        db = conf['db']
        if isinstance(db, str):
            if db[0] == '/':
                db = Path(db)
            else:
                db = Path(hutch_dir) / db
        else:
            logger.error('Invalid db conf %s, must be string.', db)
            db = None
    except KeyError:
        db = None
        logger.info(('Missing db from conf. Will skip loading from shared '
                     'database.'))
    try:
        load = conf['load']
        if not isinstance(load, (str, list)):
            logger.error('Invalid load conf %s, must be string or list', load)
            load = None
    except KeyError:
        load = None
        logger.info('Missing load from conf. Will skip loading hutch files.')

    try:
        experiment = conf['experiment']
        if not isinstance(experiment, str):
            logger.error('Invalid experiment selection %s, must be a string '
                         'matching the elog experiment name.', experiment)
            experiment = None
    except KeyError:
        experiment = None
        if hutch is None:
            logger.info(('Missing hutch and experiment from conf. Will not '
                         'load objects from questionnaire or experiment '
                         'file.'))

    try:
        # Configure whether we use the LCLS-I or LCLS-II daq
        daq_type = conf['daq_type']
    except KeyError:
        daq_type = 'lcls1'
        logger.info('Selected default daq type lcls1')
    else:
        if daq_type in ('lcls1', 'lcls1-sim', 'lcls2', 'nodaq'):
            logger.info(f'Selected valid daq type {daq_type}')
        else:
            logger.error('Selected invalid daq type! Will skip daq!')

    try:
        # Configure the daq host.
        # Required if lcls2 daq, unused if lcls1 daq.
        daq_host = conf['daq_host']
    except KeyError:
        if daq_type == 'lcls2':
            logger.error(
                'Missing required key "daq_host" in config! '
                'DAQ setup will fail!'
            )

    try:
        # This is an internal variable here for note-keeping. The ELog uses
        # this to determine if we are in the secondary or primary DAQ mode
        # In LCLS-II mode, this also sets the literal DAQ platform
        # (LCLS-I can pick it automatically)
        default_platform = True
        platform_info = conf['daq_platform']
        hostname = gethostname()
        try:
            daq_platform = platform_info[hostname]
            logger.info('Selected %s daq platform: %s',
                        hostname, daq_platform)
            default_platform = False
        except KeyError:
            daq_platform = platform_info['default']
            logger.info('Selected default %s daq platform: %s',
                        hutch, daq_platform)
    except KeyError:
        daq_platform = 0
        logger.info('Selected default hutch-python daq platform: 0')

    # Make cache namespace
    cache = LoadCache((hutch or 'hutch') + '.db', hutch_dir=hutch_dir)

    # Load debug tools
    with safe_load('debug tools'):
        load_debug(cache)

    # Load options
    with safe_load('options'):
        load_options(cache)

    # Configure ophyd
    with safe_load('configure ophyd'):
        setup_ophyd()

    cache(
        logs=HelpfulNamespace(
            log_objects=log_setup.log_objects,
            log_objects_off=log_setup.log_objects_off,
            get_log_directory=log_setup.get_log_directory,
            get_session_logfiles=log_setup.get_session_logfiles,
            get_console_level_name=log_setup.get_console_level_name,
            set_console_level=log_setup.set_console_level,
            debug_mode=log_setup.debug_mode,
            debug_context=log_setup.debug_context,
        )
    )

    # Make RunEngine
    with safe_load('run engine'):
        RE = RunEngine({})
        initialize_qt_teleporter()
        bec = BestEffortCallback()
        if matplotlib.get_backend() != 'Qt5Agg':
            logger.warning(
                'Disabling bluesky scan plots. Matplotlib config must '
                'be set up for qt5 for bluesky scans to work!'
                )
            bec.disable_plots()
        RE.subscribe(bec)
        # Enable scientific notation for big/small numbers in LiveTable
        LiveTable._FMT_MAP['number'] = 'g'
        cache(RE=RE, bec=bec)
        cache.doc(
            RE='Bluesky plan executor.',
            bec='Bluesky best-effort callback for visualization settings.',
            )

    # Collect Plans
    with safe_load('bluesky plans'):
        cache(
            bp=plan_defaults.plans,
            bps=plan_defaults.plan_stubs,
            bpp=plan_defaults.preprocessors,
            )
        cache.doc(
            bp='Namespace of full bluesky plans.',
            bps='Namespace of bluesky plan building blocks (stubs).',
            bpp='Namespace of bluesky plan preprocessors.',
            )

    # Inline calculations
    with safe_load('calc utils'):
        cache(calc=calc_defaults.calc_namespace)
        cache.doc(calc='Namespace of calculation utilities.')

    # Daq
    with safe_load('daq'):
        if daq_type.startswith('lcls1'):
            if daq_type == 'lcls1-sim' or (args and args.sim):
                set_daq_sim(True)
            cache(daq=Daq(RE=RE, hutch_name=hutch))
            cache.doc(daq='LCLS1 DAQ interface object.')
        elif daq_type == 'lcls2':
            # Warning for users of the --sim flag
            if args is not None and args.sim:
                logger.warning('Sim mode not implemented for lcls2 DAQ!')
                logger.warning('Instantiating live DAQ!')
            # Optional dependency
            from psdaq.control.DaqControl import DaqControl # NOQA
            from psdaq.control.BlueskyScan import BlueskyScan # NOQA
            daq_control = DaqControl(
                host=daq_host,
                platform=daq_platform,
                timeout=1000,
            )
            instr = daq_control.getInstrument()
            if instr is None:
                logger.error('Failed to connect to LCLS-II DAQ')
            start_state = daq_control.getState()
            if start_state == 'error':
                logger.error('DAQ is in an error state.')
            daq_bluesky = BlueskyScan(
                daq_control,
                daqState=start_state,
            )
            cache(daq=daq_bluesky)
            cache.doc(daq='LCLS2 DAQ interface object.')
        elif daq_type == 'nodaq':
            logger.info('Skip daq because daq_type is nodaq.')
        else:
            raise RuntimeError('Not loading daq, invalid daq_type.')

    # Scan PVs
    if hutch is not None:
        with safe_load('scan_pvs'):
            scan_pvs = ScanVars('{}:SCAN'.format(hutch.upper()),
                                name='scan_pvs', RE=RE)
            scan_pvs.enable()
            cache(scan_pvs=scan_pvs)
            cache.doc(scan_pvs='LCLS scan status PVs.')

    # Elog
    if hutch is not None:
        with safe_load('elog'):
            if HutchELog is None:
                txt = 'Skip elog, module not available.'
                logger.warning(txt)
                raise RuntimeError(txt)
            # Use the fact if we we used the default_platform or not to decide
            # whether we are in a specialty station or not
            if default_platform:
                logger.debug("Using primary experiment ELog")
                kwargs = dict()
            else:
                logger.info("Configuring ELog to post to secondary experiment")
                kwargs = {'station': '1'}
            cache(elog=HutchELog.from_conf(hutch.upper(), **kwargs))
            cache.doc(elog='Elog posting interface object.')

    # Shared global devices for LCLS
    with safe_load('lcls PVs'):
        cache(**global_devices())
        cache.doc(**global_device_docs)

    # Happi db and Lightpath
    if db is not None:
        with safe_load('database'):
            happi_objs = get_happi_objs(db, hutch)
            cache(**happi_objs)
            bp = get_lightpath(db, hutch)
            if bp is not None and bp.devices:
                beampath_name = "{}_beampath".format(hutch.lower())
                cache(**{beampath_name: bp})
                cache.doc(**{beampath_name: 'Lightpath beam path object.'})

    # ArchApp
    with safe_load('archapp'):
        cache(archive=EpicsArchive())
        cache.doc(archive='Archiver interface for getting historical data.')

    # Camviewer
    if hutch is not None:
        with safe_load('camviewer config'):
            objs = read_camviewer_cfg(CAMVIEWER_CFG.format(hutch))
            cache(camviewer=HelpfulNamespace(**objs))
            cache.doc(camviewer='Namespace of configured camviewer cameras.')

    # Simulated hardware
    with safe_load('simulated hardware'):
        cache(sim=sim.get_hw())
        cache.doc(sim='Namespace of simulated hardware.')

    # Auto select experiment if we need to
    if experiment is None:
        if hutch is not None:
            try:
                # xpplp1216
                experiment = get_current_experiment(hutch)
                logger.info('Selected active experiment %s', experiment)
            except Exception:
                err = 'Failed to select experiment automatically'
                logger.error(err)
                logger.debug(err, exc_info=True)

    # Process experiment name a bit
    if experiment is not None:
        if hutch in experiment:
            full_expname = experiment
            raw_expname = experiment.replace(hutch, '', 1)
        else:
            full_expname = hutch + experiment
            raw_expname = experiment

    # Load questionnaire
    if experiment is not None:
        qs_objs = get_qs_objs(full_expname)
        cache(**qs_objs)

    # Load user/beamline files
    if load is not None:
        load_objs = get_user_objs(load)
        cache(**load_objs)

    # Load experiment file
    if experiment is not None:
        user = get_exp_objs(raw_expname)
        for name, obj in qs_objs.items():
            setattr(user, name, obj)
        cache(x=user, user=user)
        cache.doc(user='User experiment object')

    # Default namespaces
    with safe_load('default groups'):
        default_class_namespace('ophyd.PositionerBase', 'motors', cache)
        default_class_namespace('Slits', 'slits', cache)
        default_class_namespace('pcdsdaq.ami.AmiDet', 'detectors', cache)

        # Hotfix/disabled until we fix issues here
        # Tree namespace can cause havoc and break top-level devices
        #
        # if hutch is not None:
        #     tree = tree_namespace(scope='hutch_python.db')
        #     # Prune meta, remove branches with only one object
        #     for name, space in tree.__dict__.items():
        #         if count_ns_leaves(space) > 1:
        #             cache(**{name: space})

        all_objs = copy(cache.objs)
        cache(a=all_objs, all_objects=all_objs)
        cache.doc(
            motors='Namespace of all positioner objects.',
            slits='Namespace of all slits objects.',
            detectors='Namespace of all ami detector objects.',
            all_objects='Namespace of all loaded objects.',
            )

    # Install Presets
    if hutch_dir is not None:
        with safe_load('position presets'):
            presets_dir = Path(hutch_dir) / 'presets'
            beamline_presets = presets_dir / 'beamline'
            preset_paths = [presets_dir, beamline_presets]
            if experiment is not None:
                experiment_presets = presets_dir / raw_expname
                preset_paths.append(experiment_presets)
            for path in preset_paths:
                if not path.exists():
                    path.mkdir()
                    path.chmod(0o777)
            if experiment is None:
                setup_preset_paths(hutch=beamline_presets)
            else:
                setup_preset_paths(hutch=beamline_presets,
                                   exp=experiment_presets)

    # Write db.txt info file to the user's module
    try:
        cache.write_file()
    except OSError:
        logger.warning('No permissions to write db.txt file')

    return cache.objs.__dict__


def default_class_namespace(cls, name, cache):
    """
    Create a class namespace and add it to the cache.

    This is an internal utility function for `load_conf.load_conf` that creates
    an `IterableNamespace`, names it, gives it a shortened name, and then
    registers both names to the `LoadCache` if the namespace isn't empty.

    Parameters
    ----------
    cls: ``type`` or ``str``
        The class to use for the namespace

    name: ``str``
        The name of the namespace

    cache: `LoadCache`
    """
    objs = class_namespace(cls, scope='hutch_python.db')
    if len(objs) > 0:
        cache(**{name: objs, name[0]: objs})



import logging
from types import SimpleNamespace

from ophyd.device import Device, Component
from ophyd.signal import Signal

from hutch_python.namespace import class_namespace, tree_namespace


logger = logging.getLogger(__name__)


def test_class_namespace():
    logger.debug('test_class_namespace')
    scope = SimpleNamespace(one=1, two=2.0, three='3', four=lambda x: 4)
    int_space = class_namespace(int, scope)
    float_space = class_namespace(float, scope)
    str_space = class_namespace(str, scope)
    func_space = class_namespace('function', scope)
    err_space = class_namespace('erqwerasd', scope)
    assert int_space.one == 1
    assert float_space.two == 2.0
    assert str_space.three == '3'
    assert func_space.four(1) == 4
    assert len(err_space) == 0


class Layer(Device):
    potato = Component(Device)


class NormalDevice(Device):
    apples = Component(Device)
    bananas = Component(Device, lazy=True)
    oranges = Component(Signal)
    veggies = Component(Layer)


def test_class_namespace_subdevices():
    logger.debug('test_class_namespace_subdevices')
    scope = SimpleNamespace(tree=NormalDevice(name='tree'))
    device_space = class_namespace(Device, scope)
    assert isinstance(device_space.tree, NormalDevice)
    assert isinstance(device_space.tree_apples, Device)
    assert isinstance(device_space.tree_veggies, Layer)
    assert isinstance(device_space.tree_veggies_potato, Device)
    assert not hasattr(device_space, 'apples')
    assert not hasattr(device_space, 'bananas')
    assert not hasattr(device_space, 'tree_bananas')
    assert not hasattr(device_space, 'oranges')
    assert not hasattr(device_space, 'tree_oranges')


def test_tree_namespace():
    logger.debug('test_metadata_namespace')
    scope = SimpleNamespace(mfx_dia_obj1=1, mfx_dia_obj2=2,
                            mfx_dg2_obj3=3, mfx_obj4=4,
                            xpp_sb2_obj5=5)
    namespaces = tree_namespace(scope=scope)
    mfx = namespaces.mfx
    xpp = namespaces.xpp
    assert mfx.dia.obj1 == 1
    assert mfx.dia.obj2 == 2
    assert mfx.dg2.obj3 == 3
    assert mfx.obj4 == 4
    assert xpp.sb2.obj5 == 5


def test_conflicting_name():
    logger.debug('test_conflicting_name')
    # This should be ok, but make sure the warning is covered
    scope = SimpleNamespace(hutch_stand=SimpleNamespace(dev=1),
                            hutch_stand_dev=2)
    tree_namespace(scope=scope)



"""
This module is used to set up and manipulate the ``logging`` configuration for
utilities like debug mode.

Functionality overview
======================

By way of :class:`hutch_python.ipython_log.IPythonLogger`, log the following
to ``{{LOG_DIR}}/year_month/user_timestamp.log``:

* All IPython input
* Any DEBUG message (well, _level 5+_)
  - Exception: hushed loggers, listed below
  - Exception: Only whitelisted ophyd object logs, down to DEBUG level (or 5)

Log to both the above file and console:
* Any INFO, WARNING, ERROR, CRITICAL messages

console exceptions:
* ophydobject INFO should be treated as DEBUG

Hush entirely - neither the file nor the console should see:
  - ophyd.event_dispatcher
  - parso
  - pyPDB.dbd.yacc
"""
import logging
import logging.config
import os
import sys
import time
from contextlib import contextmanager
from pathlib import Path
from typing import Optional, Union

import coloredlogs
import pcdsutils.log
import yaml

from . import constants
from .utils import get_fully_qualified_domain_name

logger = logging.getLogger(__name__)
central_logger = pcdsutils.log.logger
LOG_DIR = None


class LoggingNotConfiguredError(Exception):
    ...


class DefaultFormatter(logging.Formatter):
    """
    A small ``logging.Formatter`` class to patch in a default
    'ophyd_object_name' as needed to logging records.
    """

    def format(self, record):
        record.__dict__.setdefault("ophyd_object_name", "-")
        return super().format(record)


class ColoredFormatter(coloredlogs.ColoredFormatter):
    """The ``coloredlogs`` version of ``DefaultFormatter``, above."""

    def format(self, record):
        record.__dict__.setdefault("ophyd_object_name", "-")
        return super().format(record)


def get_log_filename(extension: str = '.log') -> Path:
    """
    Get a logger filename and ready it for usage.

    Parameters
    ----------
    extension : str
        The log file extension.

    Returns
    -------
    pathlib.Path :
        The log file path.
    """
    if LOG_DIR is None:
        raise LoggingNotConfiguredError(
            "Logging was not configured (LOG_DIR unset).  If in production "
            "mode, please call `configure_log_directory` first."
        )

    # Subdirectory for year/month
    dir_month = LOG_DIR / time.strftime('%Y_%m')

    # Make the log directories if they don't exist
    # Make sure each level is all permissions
    for directory in (LOG_DIR, dir_month):
        if not directory.exists():
            directory.mkdir()
            directory.chmod(0o777)

    user = os.environ['USER']
    timestamp = time.strftime('%d_%Hh%Mm%Ss')

    logfile = dir_month / f'{user}_{timestamp}{extension}'
    logfile.touch()
    return logfile


def _read_logging_config() -> dict:
    """Read the logging configuration file into a dictionary."""
    with open(constants.FILE_YAML, 'rt') as f:
        return yaml.safe_load(f.read())


def get_log_directory() -> Optional[Path]:
    """Get the currently configured logging path."""
    return LOG_DIR


def configure_log_directory(dir_logs: Optional[Union[str, Path]]):
    """
    Configure the logging path.

    Parameters
    ----------
    dir_logs: ``str`` or ``Path``, optional
        Path to the log directory. If omitted, we won't use a log file.
    """
    global LOG_DIR
    LOG_DIR = Path(dir_logs).expanduser().resolve() if dir_logs else None


def setup_logging():
    """
    Sets up the ``logging`` configuration.

    Uses ``logging.yml`` to define the config
    and manages the ``log`` directory paths.

    Also sets up the standard pcds logstash handler.
    """
    config = _read_logging_config()

    if LOG_DIR is None:
        # Remove debug file from the config
        del config['handlers']['debug']
        config['root']['handlers'].remove('debug')
    else:
        config['handlers']['debug']['filename'] = str(get_log_filename())

    # Configure centralized PCDS logging:
    fqdn = get_fully_qualified_domain_name()
    if any(fqdn.endswith(domain) for domain in constants.LOG_DOMAINS):
        pcdsutils.log.configure_pcds_logging()

    # This ensures that centralized logging messages do not make it to the
    # user or other log files.
    central_logger.propagate = False

    logging.config.dictConfig(config)
    noisy_loggers = ['ophyd.event_dispatcher', 'parso',
                     'pyPDB.dbd.yacc', 'bluesky']
    hush_noisy_loggers(noisy_loggers)


def validate_log_level(level: Union[str, int]) -> int:
    '''
    Return a logging level integer for level comparison.

    Parameters
    ----------
    level : str or int
        The logging level string or integer value.

    Returns
    -------
    log_level : int
        The integral log level.

    Raises
    ------
    ValueError
        If the logging level is invalid.
    '''
    if isinstance(level, int):
        return level
    if isinstance(level, str):
        levelno = logging.getLevelName(level)

    if not isinstance(levelno, int):
        raise ValueError("Invalid logging level (use e.g., DEBUG or 6)")

    return levelno


class ObjectFilter(logging.Filter):
    '''
    A logging filter that limits log messages to a specific object or objects.

    To be paired with a logging handler added to the ophyd object logger.

    Parameters
    ----------
    objs : iterable of OphydObject
        Ophyd objects to filter.

    level : str or int
        Cut-off level for log messages.
        Default is 'WARNING'. Python log level names or their corresponding
        integers are accepted.

    whitelist_all_level : str or int
        Cut-off level for log messages.
        Default is 'WARNING'. Python log level names or their corresponding
        integers are accepted.

    allow_other_messages : bool, optional
        Allow messages through the filter that are not specific to ophyd
        objects.  Ophyd object-related messages _must_ pass the filter.
    '''
    def __init__(self, *objects, level='WARNING',
                 whitelist_all_level='WARNING', allow_other_messages=True):
        self._objects = frozenset(objects)
        self.allow_other_messages = bool(allow_other_messages)
        self.whitelist_all_level = whitelist_all_level
        self.level = level

    def __repr__(self):
        objects = ", ".join(obj.name for obj in self.objects)
        return (
            f"{self.__class__.__name__}("
            f"level={self.level}, "
            f"allow_other_messages={self.allow_other_messages}, "
            f"objects=[{objects}]"
            f")"
        )

    def disable(self):
        """Disable the filter."""
        self.objects = []
        self.allow_other_messages = True

    @property
    def objects(self):
        """The objects to log."""
        return list(sorted(self._objects, key=lambda obj: obj.name))

    @objects.setter
    def objects(self, objects):
        self._objects = frozenset(objects)

    @property
    def levelno(self) -> int:
        """The logging level number."""
        return self._levelno

    @property
    def level(self) -> str:
        """The logging level name."""
        return logging.getLevelName(self._levelno)

    @level.setter
    def level(self, value):
        self._levelno = validate_log_level(value)

    @property
    def whitelist_all_level(self) -> str:
        """The logging level at which we whitelist *all* objects."""
        return logging.getLevelName(self._whitelist_all_levelno)

    @whitelist_all_level.setter
    def whitelist_all_level(self, value):
        self._whitelist_all_levelno = validate_log_level(value)

    @property
    def object_names(self):
        """The names of all contained objects."""
        return set(obj.name for obj in self.objects)

    def filter(self, record):
        if not hasattr(record, 'ophyd_object_name'):
            return self.allow_other_messages

        if record.levelno == logging.INFO:
            # This is rather dastardly, but we consider ophydobj INFO to be
            # closer to DEBUG.  Make it so.
            record.levelno = logging.DEBUG
            record.levelname = "DEBUG"

        if record.levelno >= self._whitelist_all_levelno:
            return True

        # if record.levelno < self.levelno:
        #     return self.allow_other_messages

        return record.ophyd_object_name in self.object_names


def find_root_object_filters():
    """
    Find all ``ObjectFilter``s configured on the root logger.

    This is useful as we configure the filters in ``logging.yaml``.

    Yields
    ------
    handler : logging.Handler
    filter : ObjectFilter
    """
    for handler in logging.root.handlers:
        for filter in handler.filters:
            if isinstance(filter, ObjectFilter):
                yield handler, filter


def log_objects_off():
    """
    Return to default logging behavior and do not treat objects specially.
    """
    log_objects(level="WARNING", console=True)


def log_objects(
    *objects,
    level: str = "DEBUG",
    console: bool = True,
):
    """
    Configure a custom logging handler on the specified object(s), and record
    log messages to files and (optionally) to the console.

    Parameters
    ----------
    *objects :
        The ``OphydObject`` instances.

    level : str, optional
        The minimum logging level to allow.  Note that the console logger
        must still have its level set appropriately. The file logger should be
        OK as it has a low default level (5, below DEBUG).

    console : bool, optional
        Apply object logging to the console as well as the rotating log file.
    """
    notification_level = 0

    old_objects = set()
    for handler, filter in find_root_object_filters():
        if isinstance(handler, logging.StreamHandler) and not console:
            continue

        old_objects = old_objects.union(filter.objects)
        filter.objects = objects
        filter.level = level
        notification_level = max((notification_level, handler.level))

    for obj in objects:
        obj.log.log(
            notification_level,
            "Recording log messages from %s (level >=%s)",
            obj.name, level
        )

    for obj in old_objects.difference(objects):
        obj.log.warning(
            "No longer recording log messages from %s",
            obj.name
        )


def hush_noisy_loggers(modules, level=logging.WARNING):
    """
    Some loggers spam on INFO with no restraint, so we must raise their levels.

    It seems there is some disagreement over what log levels should mean. In
    our repos, INFO is used as the de-facto print replacement, but in some
    repos it is used as the secondary debug stream.
    """
    for module in modules:
        logging.getLogger(module).propagate = False


def get_session_logfiles():
    """
    Get the path to the current debug log file

    Returns
    -------
    logs : list
        List of absolute paths to log files that were created by this session.
        Returns an empty list if there is no ``RotatingFileHandler`` with the
        name ``debug``
    """
    # Grab the debug file handler
    try:
        handler = get_debug_handler()
    except RuntimeError:
        logger.warning("No debug RotatingFileHandler configured for session")
        return list()
    # Find all the log files that were generated by this session
    base = Path(handler.baseFilename)
    return [str(base.parent / log)
            for log in os.listdir(base.parent)
            if log.startswith(base.name)]


def get_console_handler():
    """
    Helper function to find the console ``StreamHandler``.

    Returns
    -------
    console: ``StreamHandler``
        The ``Handler`` that prints to the screen.
    """
    return get_handler('console')


def get_debug_handler():
    """
    Helper function to find the debug ``RotatingFileHandler``

    Returns
    -------
    debug: ``RotatingFileHandler``
        The ``Handler`` that prints to the log files
    """
    return get_handler('debug')


def get_handler(name):
    """
    Helper function to get an arbitrary `Handler`

    Returns
    -------
    hander : `Handler`
    """
    root = logging.getLogger('')
    for handler in root.handlers:
        if handler.name == name:
            return handler
    raise RuntimeError('No {} handler'.format(name))


def get_console_level():
    """
    Helper function to get the console's log level.

    Returns
    -------
    level: ``int``
        Compare to ``logging.INFO``, ``logging.DEBUG``, etc. to see which log
        messages will be printed to the screen.
    """
    handler = get_console_handler()
    return handler.level


def get_console_level_name():
    """
    Helper function to get the console's log level name.

    Returns
    -------
    level: str
        The current console log level as a user-friendly string.
    """
    return logging.getLevelName(get_console_level())


def set_console_level(level=logging.INFO):
    """
    Helper function to set the console's log level.

    Parameters
    ----------
    level: int or str
        Likely one of ``logging.INFO``, ``logging.DEBUG``, etc.
    """
    handler = get_console_handler()
    handler.level = validate_log_level(level)


def debug_mode(debug=None):
    """
    Enable, disable, or check if we're in debug mode.

    Debug mode means that the console's logging level is ``logging.DEBUG`` or
    lower, which means we'll see all of the internal log messages that usually
    are not sent to the screen.

    Parameters
    ----------
    debug: ``bool``, optional
        If provided, we'll turn debug mode on (``True``) or off (``False``)

    Returns
    -------
    debug: ``bool`` or ``None``
        Returned if `debug_mode` is called with no arguments. This is ``True`
        if we're in debug mode, and ``False`` otherwise.
    """
    if debug is None:
        level = get_console_level()
        return level <= logging.DEBUG
    elif debug:
        set_console_level(level=logging.DEBUG)
    else:
        set_console_level(level=logging.INFO)


@contextmanager
def debug_context():
    """
    Context manager for running a block of code in `debug_mode`.

    For example:

    .. code-block:: python

        with debug_context():
            buggy_function()
    """
    old_level = get_console_level()
    debug_mode(True)
    yield
    set_console_level(level=old_level)


def debug_wrapper(f, *args, **kwargs):
    """
    Wrapper for running a function in `debug_mode`.

    Parameters
    ----------
    f: ``function``
        Wrapped function to call

    *args:
        Function arguments

    **kwargs:
        Function keyword arguments
    """
    with debug_context():
        f(*args, **kwargs)


def log_exception_to_central_server(
    exc_info, *,
    context='exception',
    message=None,
    level=logging.ERROR,
    stacklevel=1,
):
    """
    Log an exception to the central server (i.e., logstash/grafana).

    Parameters
    ----------
    exc_info : (exc_type, exc_value, exc_traceback)
        The exception information.

    context : str, optional
        Additional context for the log message.

    message : str, optional
        Override the default log message.

    level : int, optional
        The log level to use.  Defaults to ERROR.

    stacklevel : int, optional
        The stack level of the message being reported.  Defaults to 1,
        meaning that the message will be reported as having come from
        the caller of ``log_exception_to_central_server``.  Applies
        only to Python 3.8+, and ignored below.
    """
    exc_type, exc_value, exc_traceback = exc_info
    if issubclass(exc_type, constants.NO_LOG_EXCEPTIONS):
        return

    if not central_logger.handlers:
        # Do not allow log messages unless the central logger has been
        # configured with a log handler.  Otherwise, the log message will
        # hit the default handler and output to the terminal.
        return

    message = message or f'[{context}] {exc_value}'
    kwargs = dict()
    if sys.version_info >= (3, 8):
        kwargs = dict(stacklevel=stacklevel + 1)

    central_logger.log(level, message, exc_info=exc_info, **kwargs)



import logging
import os
import shutil
from pathlib import Path

import IPython.core.completer
import pytest
from conftest import cli_args, restore_logging

import hutch_python.cli
from hutch_python.cli import main
from hutch_python.load_conf import load

from .conftest import skip_if_win32_generic

logger = logging.getLogger(__name__)

CFG_PATH = Path(os.path.dirname(__file__)) / 'conf.yaml'
CFG = str(CFG_PATH)


@pytest.fixture(scope='function')
def no_ipython_launch(monkeypatch):
    def no_op(*args, **kwargs):
        pass
    monkeypatch.setattr(hutch_python.cli, 'start_ipython', no_op)


@skip_if_win32_generic
def test_main_normal(no_ipython_launch):
    logger.debug('test_main_normal')

    with cli_args(['hutch_python', '--cfg', CFG]):
        with restore_logging():
            main()


@skip_if_win32_generic
def test_main_no_args(no_ipython_launch):
    logger.debug('test_main_no_args')

    with cli_args(['hutch_python']):
        with restore_logging():
            main()


@skip_if_win32_generic
def test_debug_arg(no_ipython_launch):
    logger.debug('test_debug_arg')

    with cli_args(['hutch_python', '--cfg', CFG, '--debug']):
        with restore_logging():
            main()


@skip_if_win32_generic
def test_sim_arg(no_ipython_launch):
    logger.debug('test_sim_arg')

    with cli_args(['hutch_python', '--cfg', CFG, '--sim']):
        with restore_logging():
            main()


def test_create_arg():
    logger.debug('test_create_arg_dev')
    hutch = 'temp_create'
    test_dir = CFG_PATH.parent.parent.parent / hutch
    if test_dir.exists():
        shutil.rmtree(test_dir)

    with cli_args(['hutch_python', '--create', hutch]):
        with restore_logging():
            main()

    assert test_dir.exists()

    # Make sure conf.yml is valid
    load(str(test_dir / 'conf.yml'))
    shutil.rmtree(test_dir)


@skip_if_win32_generic
def test_run_script():
    logger.debug('test_run_script')

    # Will throw a name error unless we run the script inside the full env
    with cli_args(['hutch_python', '--cfg', CFG, '--debug',
                   str(Path(__file__).parent / 'script.py')]):
        with restore_logging():
            main()


def test_ipython_tab_completion():
    class MyTest:
        THIS_SHOULD_NOT_BE_THERE = None

        def __dir__(self):
            return ['foobar']

    ns = {'a': MyTest()}

    # Side-effect of the following is monkey-patching `dir2` to "fix" this for
    # us.
    hutch_python.cli.configure_ipython_session()

    completer = IPython.core.completer.Completer(namespace=ns)
    completer.limit_to__all__ = False
    assert 'a.THIS_SHOULD_NOT_BE_THERE' not in completer.attr_matches('a.')
    assert completer.attr_matches('a.') == ['a.foobar']
